# Copyright (c) 2025, Tri Dao.
# Copyright (c) 2026, Jerry Chen
"""SM90 WGMMA and shared-memory layout helpers."""

from typing import Type, Union, Optional
import cutlass
import cutlass.cute as cute
from cutlass import Int32, Float32, Boolean, const_expr
from cutlass.cute.nvgpu import warpgroup
from cutlass.cutlass_dsl import Numeric, dsl_user_op
from cutlass.utils import LayoutEnum
import cutlass.utils.hopper_helpers as sm90_utils_og


@cute.jit
def gemm(
    tiled_mma: cute.TiledMma,
    acc: cute.Tensor,
    tCrA: cute.Tensor,
    tCrB: cute.Tensor,
    zero_init: cutlass.Constexpr[bool] = False,
    wg_wait: cutlass.Constexpr[int] = 0,
    swap_AB: cutlass.Constexpr[bool] = False,
) -> None:
    if const_expr(swap_AB):
        gemm(tiled_mma, acc, tCrB, tCrA, zero_init=zero_init, wg_wait=wg_wait, swap_AB=False)
    else:
        warpgroup.fence()
        # We make a new mma_atom since we'll be modifying its attribute (accumulate).
        # Otherwise the compiler complains "operand #0 does not dominate this use"
        mma_atom = cute.make_mma_atom(tiled_mma.op)
        mma_atom.set(warpgroup.Field.ACCUMULATE, not zero_init)
        K = cute.size(tCrA.shape[2])
        # for k in cutlass.range_constexpr(K):
        # # for k in cutlass.range(K, unroll=16):
        #     cute.gemm(mma_atom, acc, tCrA[None, None, k], tCrB[None, None, k], acc)
        #     mma_atom.set(warpgroup.Field.ACCUMULATE, True)
        # for k_outer in cutlass.range_constexpr(K // 16):
        # If we unroll full, the UR is not enough for large K.
        for k_outer in cutlass.range(K // 16, unroll=1):
            for k_inner in cutlass.range_constexpr(16):
                k = k_outer * 16 + k_inner
                cute.gemm(mma_atom, acc, tCrA[None, None, k], tCrB[None, None, k], acc)
                mma_atom.set(warpgroup.Field.ACCUMULATE, True)
        for k in cutlass.range_constexpr(K - (K // 16) * 16):
            k = (K // 16) * 16 + k
            cute.gemm(mma_atom, acc, tCrA[None, None, k], tCrB[None, None, k], acc)
            mma_atom.set(warpgroup.Field.ACCUMULATE, True)
        warpgroup.commit_group()
        if const_expr(wg_wait >= 0):
            warpgroup.wait_group(wg_wait)


def gemm_zero_init(
    tiled_mma: cute.TiledMma,
    shape: cute.Shape,
    tCrA: cute.Tensor,
    tCrB: cute.Tensor,
    A_idx: Optional[Int32] = None,
    B_idx: Optional[Int32] = None,
    wg_wait: int = -1,
    swap_AB: bool = False,
) -> cute.Tensor:
    if const_expr(swap_AB):
        return gemm_zero_init(tiled_mma, shape[::-1], tCrB, tCrA, B_idx, A_idx, wg_wait, swap_AB=False)
    else:
        acc = cute.make_rmem_tensor(tiled_mma.partition_shape_C(shape), Float32)
        rA = tCrA if const_expr(A_idx is None) else tCrA[None, None, None, A_idx]
        rB = tCrB if const_expr(B_idx is None) else tCrB[None, None, None, B_idx]
        gemm(tiled_mma, acc, rA, rB, zero_init=True, wg_wait=wg_wait)
        return acc


def gemm_w_idx(
    tiled_mma: cute.TiledMma,
    acc: cute.Tensor,
    tCrA: cute.Tensor,
    tCrB: cute.Tensor,
    zero_init: Boolean,
    A_idx: Optional[Int32] = None,
    B_idx: Optional[Int32] = None,
    wg_wait: int = -1,
    swap_AB: bool = False,
) -> None:
    if const_expr(swap_AB):
        gemm_w_idx(tiled_mma, acc, tCrB, tCrA, zero_init, B_idx, A_idx, wg_wait, swap_AB=False)
    else:
        rA = tCrA if const_expr(A_idx is None) else tCrA[None, None, None, A_idx]
        rB = tCrB if const_expr(B_idx is None) else tCrB[None, None, None, B_idx]
        gemm(tiled_mma, acc, rA, rB, zero_init=zero_init, wg_wait=wg_wait)


def mma_partition_fragment_AB(
    thr_mma: cute.core.ThrMma,
    sA: Optional[cute.Tensor],
    sB: Optional[cute.Tensor],
    swap_AB: bool,
):
    if const_expr(not swap_AB):
        return (
            thr_mma.make_fragment_A(thr_mma.partition_A(sA)) if sA is not None else None,
            thr_mma.make_fragment_B(thr_mma.partition_B(sB)) if sB is not None else None,
        )
    else:
        return (
            thr_mma.make_fragment_B(thr_mma.partition_B(sA)) if sA is not None else None,
            thr_mma.make_fragment_A(thr_mma.partition_A(sB)) if sB is not None else None,
        )


@dsl_user_op
def make_smem_layout(
    dtype: Type[Numeric],
    layout: LayoutEnum,
    shape: cute.Shape,
    stage: Optional[int] = None,
    *,
    loc=None,
    ip=None,
) -> Union[cute.Layout, cute.ComposedLayout]:
    major_mode_size = shape[1] if layout.is_n_major_c() else shape[0]
    smem_layout_atom = warpgroup.make_smem_layout_atom(
        sm90_utils_og.get_smem_layout_atom(layout, dtype, major_mode_size),
        dtype,
    )
    order = (1, 0, 2) if const_expr(layout.is_m_major_c()) else (0, 1, 2)
    smem_layout_staged = cute.tile_to_shape(
        smem_layout_atom,
        cute.append(shape, stage) if const_expr(stage is not None) else shape,
        order=order if const_expr(stage is not None) else order[:2],
    )
    return smem_layout_staged



# Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
API for Grouped GEMM dSwiGLU Backward Kernel (SM100+)

This module provides the API class for contiguous grouped block-scaled GEMM
backward pass with dSwiGLU activation gradient for MoE (Mixture of Experts) workloads.
"""

from .grouped_gemm_dswiglu_quant import (
    BlockScaledContiguousGroupedGemmKernel,
)
from cuda.bindings import driver as cuda
import torch
from typing import Tuple, Optional

import cutlass
import cutlass.cute as cute
from cutlass.cute.runtime import make_fake_stream

from cudnn.datatypes import _convert_to_cutlass_data_type
from cudnn.api_base import APIBase, TupleDict, ceil_div, is_power_of_2


class GroupedGemmDswigluSm100(APIBase):
    """API class for Grouped GEMM dSwiGLU backward operation on SM100+ GPUs.

    This kernel performs contiguous grouped block-scaled GEMM backward pass
    with dSwiGLU activation gradient, designed for MoE (Mixture of Experts) workloads.

    Key features:
    - Supports variable M per group (aligned to m_aligned)
    - Contiguous memory layout for A and D tensors
    - Block-scaled quantization support (MXF8, MXF4, NVF4)
    - Uses padded_offsets interface for expert mapping

    Example:
        >>> api = GroupedGemmDswigluSm100(
        ...     sample_a=a_tensor,
        ...     ...
        ... )
        >>> api.check_support()
        >>> api.compile()
        >>> api.execute(..., stream)
    """

    def __init__(
        self,
        sample_a: torch.Tensor,
        sample_b: torch.Tensor,
        sample_c: torch.Tensor,
        sample_d_row: torch.Tensor,
        sample_d_col: torch.Tensor,
        sample_sfa: torch.Tensor,
        sample_sfb: torch.Tensor,
        sample_padded_offsets: torch.Tensor,
        sample_alpha: torch.Tensor,
        sample_beta: torch.Tensor,
        sample_prob: torch.Tensor,
        sample_dprob: torch.Tensor,
        # Optional quantization output arguments
        sample_sfd_row: Optional[torch.Tensor] = None,
        sample_sfd_col: Optional[torch.Tensor] = None,
        sample_amax: Optional[torch.Tensor] = None,
        sample_norm_const: Optional[torch.Tensor] = None,
        # Configuration
        acc_dtype: torch.dtype = torch.float32,
        mma_tiler_mn: Tuple[int, int] = (256, 256),
        cluster_shape_mn: Optional[Tuple[int, int]] = None,
        sf_vec_size: int = 16,
        vector_f32: bool = False,
        m_aligned: int = 256,
        discrete_col_sfd: bool = False,
        epilogue_op: Optional[str] = None,
    ):
        """Initialize the GroupedGemmDswigluSm100 API.

        :param sample_a: Sample A tensor (valid_m, k, 1)
        :param sample_b: Sample B tensor (n, k, l) where l = num_groups
        :param sample_c: Sample C tensor for intermediate storage (valid_m, 2n, 1)
        :param sample_d_row: Sample D row output tensor (valid_m, 2n, 1) after dSwiGLU
        :param sample_d_col: Sample D column output tensor (valid_m, 2n, 1) after dSwiGLU
        :param sample_sfa: Sample scale factor A tensor
        :param sample_sfb: Sample scale factor B tensor
        :param sample_padded_offsets: End offset for each expert after padding, shape (expert_cnt,)
        :param sample_alpha: Per-group alpha scaling factors
        :param sample_beta: Per-group beta scaling factors
        :param sample_prob: Per-row probability tensor (valid_m, 1, 1)
        :param sample_dprob: Gradient of probability tensor (valid_m, 1, 1). Must be zero-initialized.
        :param sample_sfd_row: Optional row scale factor for D
        :param sample_sfd_col: Optional column scale factor for D
        :param sample_amax: Optional amax tensor for quantization, shape (l, 2, 1)
        :param sample_norm_const: Optional normalization constant
        :param acc_dtype: Accumulator data type
        :param mma_tiler_mn: MMA tiler shape (M, N)
        :param cluster_shape_mn: Cluster shape (M, N)
        :param sf_vec_size: Scale factor vector size
        :param vector_f32: Use vectorized f32 operations
        :param m_aligned: Alignment for group M dimension
        :param discrete_col_sfd: Boolean, True to generate discrete col-major scale factor tensor
        :param epilogue_op: Optional epilogue operation. Valid values: None, "none", "identity", "relu", "srelu"
        """
        super().__init__()

        self._warn_experimental_api()
        self._logger.debug("Entering __init__")

        # Store sample tensor descriptors
        self.a_desc = self._make_tensor_desc(sample_a, name="sample_a")
        self.b_desc = self._make_tensor_desc(sample_b, name="sample_b")
        self.c_desc = self._make_tensor_desc(sample_c, name="sample_c")
        self.d_row_desc = self._make_tensor_desc(sample_d_row, name="sample_d_row")
        self.d_col_desc = self._make_tensor_desc(sample_d_col, name="sample_d_col")
        self.sfa_desc = self._make_tensor_desc(sample_sfa, name="sample_sfa")
        self.sfb_desc = self._make_tensor_desc(sample_sfb, name="sample_sfb")
        self.padded_offsets_desc = self._make_tensor_desc(sample_padded_offsets, name="sample_padded_offsets")
        self.alpha_desc = self._make_tensor_desc(sample_alpha, name="sample_alpha")
        self.beta_desc = self._make_tensor_desc(sample_beta, name="sample_beta")
        self.prob_desc = self._make_tensor_desc(sample_prob, name="sample_prob")
        self.dprob_desc = self._make_tensor_desc(sample_dprob, name="sample_dprob")

        # Optional quantization outputs
        self.sfd_row_desc = self._make_tensor_desc(sample_sfd_row, name="sample_sfd_row")
        self.sfd_col_desc = self._make_tensor_desc(sample_sfd_col, name="sample_sfd_col")
        self.amax_desc = self._make_tensor_desc(sample_amax, name="sample_amax")
        self.norm_const_desc = self._unpad_tensor_to_ndim(self._make_tensor_desc(sample_norm_const, name="sample_norm_const"), 1, "norm_const")

        # Configuration
        self.acc_dtype = acc_dtype
        self.mma_tiler_mn = mma_tiler_mn
        self.use_2cta_instrs = mma_tiler_mn[0] == 256
        if cluster_shape_mn is None:
            self.cluster_shape_mn = (2, 1) if self.use_2cta_instrs else (1, 1)
        else:
            self.cluster_shape_mn = cluster_shape_mn
        self.sf_vec_size = sf_vec_size
        self.vector_f32 = vector_f32
        self.m_aligned = m_aligned
        self.discrete_col_sfd = discrete_col_sfd

        # expert_cnt derived from padded_offsets shape
        self.expert_cnt = self.padded_offsets_desc.shape[0]

        # Epilogue operation
        if epilogue_op in [None, "none", "identity"]:
            self.epilogue_op = lambda x: x
        elif epilogue_op == "relu":
            self.epilogue_op = lambda x: cute.where(x > 0, x, cute.full_like(x, 0))
        elif epilogue_op == "srelu":
            self.epilogue_op = lambda x: cute.where(x > 0, x, cute.full_like(x, 0)) ** 2
        else:
            raise ValueError(f"Invalid epilogue operation: {epilogue_op}. Valid values: None, 'none', 'identity', 'relu', 'srelu'")

        self._interpret_uint8_as_fp4x2 = True
        self._kernel = BlockScaledContiguousGroupedGemmKernel
        self.num_cluster_overlap_margin = int(os.getenv("CUDNNFE_CLUSTER_OVERLAP_MARGIN", "0"))
        self._logger.debug(f"setting num_cluster_overlap_margin: {self.num_cluster_overlap_margin}")
        self._logger.debug(f"__init__ completed")

    def check_support(self) -> bool:
        """Check if the kernel configuration is supported.

        :return: True if supported, raises exception otherwise
        """
        self._logger.debug("Entering check_support")

        all_none = all(x is None for x in [self.sfd_row_desc, self.sfd_col_desc, self.norm_const_desc])
        none_none = all(x is not None for x in [self.sfd_row_desc, self.sfd_col_desc, self.norm_const_desc])
        if not (all_none or none_none):
            raise ValueError("sfd_row, sfd_col, and norm_const must be all None or all not None")
        self.generate_sfd = none_none
        if self.discrete_col_sfd and not self.generate_sfd:
            self._logger.warning("discrete_col_sfd is True but generate_sfd is False, discrete_col_sfd will be ignored")
            self.discrete_col_sfd = False

        self._logger.debug("Checking tensor shapes and strides")
        tensor_m, k, _one = self.a_desc.shape
        n, _, l = self.b_desc.shape
        _, _, _one = self.c_desc.shape
        self._check_tensor_shape(self.a_desc, (tensor_m, k, 1), "A")
        self._check_tensor_shape(self.b_desc, (n, k, l), "B")
        self._check_tensor_shape(self.c_desc, (tensor_m, n * 2, 1), "C")
        self._check_tensor_shape(self.d_row_desc, (tensor_m, n * 2, 1), "D_row")
        self._check_tensor_shape(self.d_col_desc, (tensor_m, n * 2, 1), "D_col")

        rest_k = ceil_div(ceil_div(k, self.sf_vec_size), 4)
        self._check_tensor_shape(self.sfa_desc, (32, 4, ceil_div(tensor_m, 128), 4, rest_k, 1), "SFA")
        self._check_tensor_shape(self.sfb_desc, (32, 4, ceil_div(n, 128), 4, rest_k, l), "SFB")
        # SFD uses full n dimension since D has n columns (interleaved ab and dswiglu)
        rest_n2_full = ceil_div(ceil_div(n * 2, self.sf_vec_size), 4)
        self._check_tensor_shape(self.sfd_row_desc, (32, 4, ceil_div(tensor_m, 128), 4, rest_n2_full, 1), "SFD_row")
        rest_m = ceil_div(ceil_div(tensor_m, self.sf_vec_size), 4)
        self._check_tensor_shape(self.sfd_col_desc, (32, 4, ceil_div(n * 2, 128), 4, rest_m, 1), "SFD_col")

        self._check_tensor_shape(self.padded_offsets_desc, (l,), "padded_offsets")
        self._check_tensor_shape(self.alpha_desc, (l,), "alpha")
        self._check_tensor_shape(self.beta_desc, (l,), "beta")
        self._check_tensor_shape(self.prob_desc, (tensor_m, 1, 1), "prob")
        self._check_tensor_shape(self.dprob_desc, (tensor_m, 1, 1), "dprob")
        self._check_tensor_shape(self.amax_desc, (l, 2, 1), "amax")
        self._check_tensor_shape(self.norm_const_desc, (1,), "norm_const")

        _ = self._check_tensor_stride(self.a_desc, stride=[(k, 1, tensor_m * k)], extra_error_msg="A must have k-major layout")
        if self._is_fp8(self.a_desc):
            _ = self._check_tensor_stride(
                self.b_desc, stride=[(k, 1, n * k), (1, n, n * k)], extra_error_msg="For fp8 ab_dtype, B must have k- or n-major layout"
            )
        else:
            _ = self._check_tensor_stride(self.b_desc, stride=[(k, 1, n * k)], extra_error_msg="For fp4 ab_dtype, B must have k-major layout")
        _ = self._check_tensor_stride(self.c_desc, stride=[(n * 2, 1, tensor_m * n * 2)], extra_error_msg="C must have n-major layout")
        # D has same shape as C (n columns for interleaved ab and dswiglu)
        _ = self._check_tensor_stride(self.d_row_desc, stride=[(n * 2, 1, tensor_m * n * 2)], extra_error_msg="D_row must have n-major layout")
        _ = self._check_tensor_stride(self.d_col_desc, stride=[(n * 2, 1, tensor_m * n * 2)], extra_error_msg="D_col must have n-major layout")

        self._logger.debug("Checking data types")
        self.ab_dtype = self._check_dtype(
            self.a_desc,
            dtype=[
                torch.float4_e2m1fn_x2,
                torch.uint8,
                torch.float8_e5m2,
                torch.float8_e4m3fn,
            ],
            name="A/B",
        )
        self._check_dtype(self.b_desc, dtype=self.ab_dtype, name="B", extra_error_msg="B must have the same dtype as A")

        self.sf_dtype = self._check_dtype(
            self.sfa_desc,
            dtype=[torch.float8_e8m0fnu, torch.float8_e4m3fn],
            name="SFA/SFB/SFD_row/SFD_col",
        )
        self._check_dtype(self.sfb_desc, dtype=self.sf_dtype, name="SFB", extra_error_msg="SFB must have the same dtype as SFA")
        self._check_dtype(self.sfd_row_desc, dtype=self.sf_dtype, name="SFD_row", extra_error_msg="SFD_row must have the same dtype as SFA")
        self._check_dtype(self.sfd_col_desc, dtype=self.sf_dtype, name="SFD_col", extra_error_msg="SFD_col must have the same dtype as SFA")

        if self.sf_vec_size not in [16, 32]:
            raise ValueError(f"sf_vec_size must be 16 or 32, got {self.sf_vec_size}")
        if self.sf_dtype in [torch.float8_e4m3fn] and self.sf_vec_size == 32:
            raise ValueError(f"sf_dtype {self.sf_dtype} and sf_vec_size {self.sf_vec_size} combination is not supported")
        if self._is_fp8(self.ab_dtype) and self.sf_vec_size == 16:
            raise ValueError(f"ab_dtype {self.ab_dtype} and sf_vec_size {self.sf_vec_size} combination is not supported")

        self._check_dtype(self.acc_dtype, dtype=torch.float32, name="Accumulator", extra_error_msg="Accumulator must be float32")
        self._check_dtype(self.prob_desc, dtype=torch.float32, name="Prob", extra_error_msg="Prob must be float32")
        self._check_dtype(self.dprob_desc, dtype=torch.float32, name="Dprob", extra_error_msg="Dprob must be float32")
        self.c_dtype = self._check_dtype(
            self.c_desc,
            dtype=[torch.float32, torch.float16, torch.bfloat16, torch.float8_e4m3fn, torch.float8_e5m2],
            name="C",
        )
        if self._is_fp8(self.c_dtype) and self.vector_f32:
            raise ValueError(
                f"Invalid configuration: fp8 ab_dtype, c_dtype, and vector_f32 is not supported. Please use vector_f32=False or c_dtype=bfloat16 instead"
            )

        if self._is_fp4x2(self.ab_dtype):
            self.d_dtype = self._check_dtype(
                self.d_row_desc,
                dtype=[torch.float16, torch.bfloat16, torch.float32],
                name="D_row",
                extra_error_msg="D_row must be fp16, bf16, or float32 when ab_dtype is fp4",
            )
        elif self._is_fp8(self.ab_dtype):
            self.d_dtype = self._check_dtype(
                self.d_row_desc,
                dtype=[
                    torch.float8_e4m3fn,
                    torch.float8_e5m2,
                ],
                name="D_row",
                extra_error_msg="D_row must be fp8 dtype when ab_dtype is fp8",
            )
        else:
            raise NotImplementedError(f"Invalid ab_dtype: {self.ab_dtype}, expected fp4 or fp8")
        self._check_dtype(self.d_col_desc, dtype=self.d_dtype, name="D_col", extra_error_msg="D_col must have the same dtype as D_row")

        self._logger.debug("Checking MMA tile shape and cluster shape")
        if not self.use_2cta_instrs and self.mma_tiler_mn[0] not in [64, 128]:
            raise ValueError(f"MMA tiler M must be 64 or 128 when use_2cta_instrs=False, got {self.mma_tiler_mn[0]}")
        if self.use_2cta_instrs and self.mma_tiler_mn[0] not in [128, 256]:
            raise ValueError(f"MMA tiler M must be 128 or 256 when use_2cta_instrs=True, got {self.mma_tiler_mn[0]}")
        if self.mma_tiler_mn[1] not in [128, 256]:
            raise ValueError(f"MMA tiler N must be 128 or 256, got {self.mma_tiler_mn[1]}")
        if self.cluster_shape_mn[0] % (2 if self.use_2cta_instrs else 1) != 0:
            raise ValueError(f"cluster_shape_mn[0] must be divisible by 2 when use_2cta_instrs=True, got {self.cluster_shape_mn[0]}")
        if not (
            self.cluster_shape_mn[0] * self.cluster_shape_mn[1] <= 16
            and self.cluster_shape_mn[0] > 0
            and self.cluster_shape_mn[1] > 0
            and self.cluster_shape_mn[0] <= 4
            and self.cluster_shape_mn[1] <= 4
            and is_power_of_2(self.cluster_shape_mn[0])
            and is_power_of_2(self.cluster_shape_mn[1])
        ):
            raise ValueError(
                f"Invalid cluster shape: expected values to be powers of 2 and cluster_shape_mn[0] * cluster_shape_mn[1] <= 16, got {self.cluster_shape_mn[0]},{self.cluster_shape_mn[1]}"
            )
        cluster_tiler_m = (self.cluster_shape_mn[0] // (2 if self.use_2cta_instrs else 1)) * self.mma_tiler_mn[0]
        if cluster_tiler_m not in [128, 256]:
            raise ValueError(f"Invalid cluster tiler shape: expected cluster_tiler_m in {{128, 256}}, got {cluster_tiler_m}")
        if self.m_aligned % self.mma_tiler_mn[0] != 0:
            raise ValueError(f"Invalid m_aligned: expected m_aligned to be divisible by mma_tiler_mn[0], got {self.m_aligned} % {self.mma_tiler_mn[0]} != 0")
        if self.m_aligned != BlockScaledContiguousGroupedGemmKernel.FIX_PAD_SIZE:
            raise ValueError(
                f"Invalid m_aligned: expected m_aligned to be equal to FIX_PAD_SIZE, got {self.m_aligned} != {BlockScaledContiguousGroupedGemmKernel.FIX_PAD_SIZE}"
            )

        self._logger.debug("Checking tensor alignment")

        def check_contigous_16B_alignment(dtype, stride_order, tensor_shape):
            is_mode0_major = stride_order == (0, 1, 2)
            major_mode_idx = 0 if is_mode0_major else 1
            num_major_elements = tensor_shape[major_mode_idx]
            num_contiguous_elements = 16 * 8 // (_convert_to_cutlass_data_type(dtype, interpret_uint8_as_fp4x2=self._interpret_uint8_as_fp4x2).width)
            return num_major_elements % num_contiguous_elements == 0

        if not (
            check_contigous_16B_alignment(self.ab_dtype, self.a_desc.stride_order, (tensor_m, k, l))
            and check_contigous_16B_alignment(self.ab_dtype, self.b_desc.stride_order, (n, k, l))
            and check_contigous_16B_alignment(self.d_dtype, self.d_row_desc.stride_order, (tensor_m, n, l))  # c, d_row, and d_col have the same stride order
        ):
            raise ValueError("Invalid tensor alignment: tensors must be 16B aligned")

        # Check expert_cnt constraint
        if self.expert_cnt > 1024:
            raise ValueError(f"expert_cnt must be <= 1024, got {self.expert_cnt}")

        # Check environment
        if not torch.cuda.is_available():
            raise RuntimeError("CUDA is not available")
        device = torch.cuda.current_device()
        major, minor = torch.cuda.get_device_capability(device)
        compute_capability = major * 10 + minor
        if compute_capability < 100:
            raise RuntimeError(f"GroupedGemmDswiglu requires SM100+ compute capability, " f"but found SM{compute_capability} on device {device}")

        self._is_supported = True
        self._logger.debug("check_support completed successfully")
        return True

    def compile(self) -> None:
        """Compile the kernel."""
        self._logger.debug("Entering compile")
        self._ensure_support_checked()
        if self._compiled_kernel is not None:
            self._logger.debug("Kernel already compiled; skipping recompilation")
            return

        gemm_dswiglu = self._kernel(
            sf_vec_size=self.sf_vec_size,
            acc_dtype=_convert_to_cutlass_data_type(self.acc_dtype),
            use_2cta_instrs=self.use_2cta_instrs,
            mma_tiler_mn=self.mma_tiler_mn,
            cluster_shape_mn=self.cluster_shape_mn,
            vectorized_f32=self.vector_f32,
            discrete_col_sfd=self.discrete_col_sfd,
            expert_cnt=self.expert_cnt,
            use_mono_increase_expert_idx=True,
        )

        hardware_info = cutlass.utils.HardwareInfo()
        max_active_clusters = hardware_info.get_max_active_clusters(self.cluster_shape_mn[0] * self.cluster_shape_mn[1])
        max_active_clusters -= self.num_cluster_overlap_margin
        self._value_error_if(
            max_active_clusters <= 0,
            "max_active_clusters must be > 0 after applying overlap margin; reduce CUDNNFE_CLUSTER_OVERLAP_MARGIN",
        )
        fake_stream = make_fake_stream(use_tvm_ffi_env_stream=False)

        self._logger.debug("Compiling grouped_gemm_dswiglu kernel")
        use_full_dynamic = os.environ.get("CUDNN_FE_GROUPED_GEMM_DYNAMIC_MNKL", "1") != "0"
        if not use_full_dynamic:
            valid_m = cute.sym_int(divisibility=256)

            a_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.a_desc.dtype,
                shape=(valid_m, *self.a_desc.shape[1:]),
                stride_order=self.a_desc.stride_order,
            )
            b_cute_fake = self._make_fake_cute_tensor_from_desc(self.b_desc, assumed_align=16)
            c_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.c_desc.dtype,
                shape=(valid_m, *self.c_desc.shape[1:]),
                stride_order=self.c_desc.stride_order,
            )
            d_row_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.d_row_desc.dtype,
                shape=(valid_m, *self.d_row_desc.shape[1:]),
                stride_order=self.d_row_desc.stride_order,
            )
            d_col_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.d_col_desc.dtype,
                shape=(valid_m, *self.d_col_desc.shape[1:]),
                stride_order=self.d_col_desc.stride_order,
            )
            prob_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.prob_desc.dtype,
                shape=(valid_m, 1, 1),
                stride_order=self.prob_desc.stride_order,
            )
            dprob_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.dprob_desc.dtype,
                shape=(valid_m, 1, 1),
                stride_order=self.dprob_desc.stride_order,
            )

            tensor_m_128 = cute.sym_int()
            stride_tensor_m_128 = cute.sym_int(divisibility=32 * 4 * 4)
            sfa_cute_fake = self._make_fake_cute_tensor(
                dtype=self.sfa_desc.dtype,
                shape=(32, 4, tensor_m_128, 4, self.sfa_desc.shape[4], 1),
                stride=(16, 4, self.sfa_desc.stride[2], 1, 512, stride_tensor_m_128),
            )
            sfb_cute_fake = self._make_fake_cute_tensor_from_desc(self.sfb_desc, assumed_align=16)

            sfd_row_fake = None
            sfd_col_fake = None
            if self.sfd_row_desc is not None:
                stride_sfd_m = cute.sym_int(divisibility=32 * 4 * 4)
                sfd_row_fake = self._make_fake_cute_tensor(
                    dtype=self.sfd_row_desc.dtype,
                    shape=(32, 4, tensor_m_128, 4, self.sfd_row_desc.shape[4], 1),
                    stride=(16, 4, self.sfd_row_desc.stride[2], 1, 512, stride_sfd_m),
                )
            if self.sfd_col_desc is not None:
                rest_m = cute.sym_int(divisibility=1)
                stride_sfd_n = cute.sym_int(divisibility=32 * 4 * 4)
                stride_rest_m = cute.sym_int(divisibility=32 * 4 * 4)
                sfd_col_fake = self._make_fake_cute_tensor(
                    dtype=self.sfd_col_desc.dtype,
                    shape=(32, 4, self.sfd_col_desc.shape[2], 4, rest_m, 1),
                    stride=(16, 4, stride_rest_m, 1, 512, stride_sfd_n),
                )
        else:
            valid_m = cute.sym_int(divisibility=256)
            n_2 = cute.sym_int()
            k = cute.sym_int()
            l = cute.sym_int()

            a_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.a_desc.dtype,
                shape=(valid_m, k, 1),
                stride_order=self.a_desc.stride_order,
                dynamic_mode=self.a_desc.stride_order[0],
                divisibility=32 if self._is_fp4x2(self.ab_dtype) else 16,
            )
            b_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.b_desc.dtype,
                shape=(cute.sym_int(), k, l),
                stride_order=self.b_desc.stride_order,
                dynamic_mode=self.b_desc.stride_order[0],
                divisibility=32 if self._is_fp4x2(self.ab_dtype) else 16,
            )
            c_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.c_desc.dtype,
                shape=(valid_m, n_2, 1),
                stride_order=self.c_desc.stride_order,
                dynamic_mode=self.c_desc.stride_order[0],
                divisibility=8 if self._is_f16(self.c_desc.dtype) else 16,
            )
            d_row_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.d_row_desc.dtype,
                shape=(valid_m, n_2, 1),
                stride_order=self.d_row_desc.stride_order,
                dynamic_mode=self.d_row_desc.stride_order[0],
                divisibility=8 if self._is_f16(self.d_row_desc.dtype) else 16,
            )
            d_col_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.d_col_desc.dtype,
                shape=(valid_m, n_2, 1),
                stride_order=self.d_col_desc.stride_order,
                dynamic_mode=self.d_col_desc.stride_order[0],
                divisibility=8 if self._is_f16(self.d_col_desc.dtype) else 16,
            )
            prob_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.prob_desc.dtype,
                shape=(valid_m, 1, 1),
                stride_order=self.prob_desc.stride_order,
            )
            dprob_cute_fake = self._make_fake_cute_compact_tensor(
                dtype=self.dprob_desc.dtype,
                shape=(valid_m, 1, 1),
                stride_order=self.dprob_desc.stride_order,
            )

            tensor_m_128 = cute.sym_int()
            rest_k = cute.sym_int()
            stride_rest_k = cute.sym_int(divisibility=32 * 4 * 4)
            stride_tensor_m_128 = cute.sym_int(divisibility=32 * 4 * 4)
            sfa_cute_fake = self._make_fake_cute_tensor(
                dtype=self.sfa_desc.dtype,
                shape=(32, 4, tensor_m_128, 4, rest_k, 1),
                stride=(16, 4, stride_rest_k, 1, 512, stride_tensor_m_128),
            )
            tensor_n_128 = cute.sym_int()
            stride_sfb_rest_k = cute.sym_int(divisibility=32 * 4 * 4)
            stride_sfb_tensor_n_128 = cute.sym_int(divisibility=32 * 4 * 4)
            sfb_cute_fake = self._make_fake_cute_tensor(
                dtype=self.sfb_desc.dtype,
                shape=(32, 4, tensor_n_128, 4, rest_k, l),
                stride=(16, 4, stride_sfb_tensor_n_128, 1, 512, stride_sfb_rest_k),
            )

            sfd_row_fake = None
            sfd_col_fake = None
            if self.sfd_row_desc is not None:
                rest_n2 = cute.sym_int()
                stride_sfd_rest_n2 = cute.sym_int(divisibility=32 * 4 * 4)
                stride_sfd_rest_tensor_m_128 = cute.sym_int(divisibility=32 * 4 * 4)
                sfd_row_fake = self._make_fake_cute_tensor(
                    dtype=self.sfd_row_desc.dtype,
                    shape=(32, 4, tensor_m_128, 4, rest_n2, 1),
                    stride=(16, 4, stride_sfd_rest_n2, 1, 512, stride_sfd_rest_tensor_m_128),
                )
            if self.sfd_col_desc is not None:
                tensor_n2_128 = cute.sym_int()
                rest_m = cute.sym_int()
                stride_sfd_rest_m = cute.sym_int(divisibility=32 * 4 * 4)
                stride_sfd_n2 = cute.sym_int(divisibility=32 * 4 * 4)
                sfd_col_fake = self._make_fake_cute_tensor(
                    dtype=self.sfd_col_desc.dtype,
                    shape=(32, 4, tensor_n2_128, 4, rest_m, 1),
                    stride=(16, 4, stride_sfd_rest_m, 1, 512, stride_sfd_n2),
                )

        _compiled_kernel = cute.compile(
            gemm_dswiglu,
            a=a_cute_fake,
            b=b_cute_fake,
            c=c_cute_fake,
            d=d_row_cute_fake,
            d_col=d_col_cute_fake,
            sfa=sfa_cute_fake,
            sfb=sfb_cute_fake,
            sfd_row_tensor=sfd_row_fake,
            sfd_col_tensor=sfd_col_fake,
            amax_tensor=self._make_fake_cute_tensor_from_desc(self.amax_desc, assumed_align=16),
            norm_const_tensor=self._make_fake_cute_tensor_from_desc(self.norm_const_desc, assumed_align=16),
            padded_offsets=self._make_fake_cute_tensor_from_desc(self.padded_offsets_desc, assumed_align=16),
            alpha=self._make_fake_cute_tensor_from_desc(self.alpha_desc, assumed_align=16),
            beta=self._make_fake_cute_tensor_from_desc(self.beta_desc, assumed_align=16),
            prob=prob_cute_fake,
            dprob=dprob_cute_fake,
            max_active_clusters=max_active_clusters,
            epilogue_op=self.epilogue_op,
            stream=fake_stream,
            options="--enable-tvm-ffi",
        )

        def tensor_api(
            a_tensor: torch.Tensor,
            b_tensor: torch.Tensor,
            c_tensor: torch.Tensor,
            d_row_tensor: torch.Tensor,
            d_col_tensor: Optional[torch.Tensor],
            sfa_tensor: torch.Tensor,
            sfb_tensor: torch.Tensor,
            sfd_row_tensor: Optional[torch.Tensor],
            sfd_col_tensor: Optional[torch.Tensor],
            amax_tensor: Optional[torch.Tensor],
            norm_const_tensor: Optional[torch.Tensor],
            padded_offsets: torch.Tensor,
            alpha_tensor: torch.Tensor,
            beta_tensor: torch.Tensor,
            prob_tensor: torch.Tensor,
            dprob_tensor: torch.Tensor,
            stream: cuda.CUstream,
        ) -> None:
            norm_const_tensor = self._unpad_tensor_to_ndim(norm_const_tensor, 1, "norm_const")
            _compiled_kernel(
                a_tensor,
                b_tensor,
                c_tensor,
                d_row_tensor,
                d_col_tensor,
                sfa_tensor,
                sfb_tensor,
                sfd_row_tensor,
                sfd_col_tensor,
                amax_tensor,
                norm_const_tensor,
                padded_offsets,
                alpha_tensor,
                beta_tensor,
                prob_tensor,
                dprob_tensor,
                stream,
            )

        self._compiled_kernel = tensor_api

        self._logger.debug("Kernel compiled successfully")

    def execute(
        self,
        a_tensor: torch.Tensor,
        b_tensor: torch.Tensor,
        c_tensor: torch.Tensor,
        d_row_tensor: torch.Tensor,
        d_col_tensor: torch.Tensor,
        sfa_tensor: torch.Tensor,
        sfb_tensor: torch.Tensor,
        padded_offsets: torch.Tensor,
        alpha_tensor: torch.Tensor,
        beta_tensor: torch.Tensor,
        prob_tensor: torch.Tensor,
        dprob_tensor: torch.Tensor,
        sfd_row_tensor: Optional[torch.Tensor] = None,
        sfd_col_tensor: Optional[torch.Tensor] = None,
        amax_tensor: Optional[torch.Tensor] = None,
        norm_const_tensor: Optional[torch.Tensor] = None,
        current_stream: Optional[cuda.CUstream] = None,
    ) -> None:
        """Execute the compiled kernel.

        :param a_tensor: Input A tensor
        :param b_tensor: Input B tensor (weights)
        :param c_tensor: Intermediate C tensor (from forward pass)
        :param d_row_tensor: Output D row tensor
        :param d_col_tensor: Output D column tensor
        :param sfa_tensor: Scale factor A
        :param sfb_tensor: Scale factor B
        :param padded_offsets: End offset per expert after padding
        :param alpha_tensor: Per-group alpha scaling factors
        :param beta_tensor: Per-group beta scaling factors
        :param prob_tensor: Per-row probability tensor
        :param dprob_tensor: Gradient of probability tensor. Must be zero-initialized.
        :param sfd_row_tensor: Optional row scale factor D
        :param sfd_col_tensor: Optional column scale factor D
        :param amax_tensor: Optional amax tensor
        :param norm_const_tensor: Optional normalization constant
        :param current_stream: CUDA stream
        """
        self._logger.debug("Entering execute")
        current_stream = self._get_default_stream(current_stream)

        if self._compiled_kernel is None:
            raise RuntimeError("Kernel not compiled; call compile() first")
        self._logger.debug("Executing grouped_gemm_dswiglu kernel")
        self._compiled_kernel(
            a_tensor=a_tensor,
            b_tensor=b_tensor,
            c_tensor=c_tensor,
            d_row_tensor=d_row_tensor,
            d_col_tensor=d_col_tensor,
            sfa_tensor=sfa_tensor,
            sfb_tensor=sfb_tensor,
            sfd_row_tensor=sfd_row_tensor,
            sfd_col_tensor=sfd_col_tensor,
            amax_tensor=amax_tensor,
            norm_const_tensor=norm_const_tensor,
            padded_offsets=padded_offsets,
            alpha_tensor=alpha_tensor,
            beta_tensor=beta_tensor,
            prob_tensor=prob_tensor,
            dprob_tensor=dprob_tensor,
            stream=current_stream,
        )

        self._logger.debug("Execute completed")


import logging
import os

_logger = logging.getLogger(__name__)
_cache_of_GroupedGemmDswigluSm100Objects = {}


def grouped_gemm_dswiglu_wrapper_sm100(
    a_tensor: torch.Tensor,
    b_tensor: torch.Tensor,
    c_tensor: torch.Tensor,  # Intermediate from forward pass (required)
    sfa_tensor: torch.Tensor,
    sfb_tensor: torch.Tensor,
    padded_offsets: torch.Tensor,
    alpha_tensor: torch.Tensor,
    beta_tensor: Optional[torch.Tensor],
    prob_tensor: torch.Tensor,
    norm_const_tensor: Optional[torch.Tensor] = None,
    acc_dtype: torch.dtype = torch.float32,
    d_dtype: torch.dtype = torch.bfloat16,
    cd_major: str = "n",
    mma_tiler_mn: Tuple[int, int] = (256, 256),
    cluster_shape_mn: Optional[Tuple[int, int]] = None,
    sf_vec_size: int = 16,
    vector_f32: bool = False,
    m_aligned: int = 256,
    discrete_col_sfd: bool = False,
    epilogue_op: Optional[str] = None,
    current_stream: Optional[cuda.CUstream] = None,
    dprob_tensor_buf: Optional[torch.Tensor] = None,
    amax_tensor_buf: Optional[torch.Tensor] = None,
) -> TupleDict:
    """Convenience wrapper for grouped GEMM dSwiGLU backward operation.

    This function creates the API, compiles, and executes in one call.
    Compiled kernels are cached for reuse when called with the same configuration.

    Args:
        a_tensor: Input A tensor (valid_m, k, 1)
        b_tensor: Weight B tensor (n, k, l)
        c_tensor: Intermediate C tensor from forward pass (valid_m, 2n, 1)
        sfa_tensor: Scale factor A
        sfb_tensor: Scale factor B
        padded_offsets: End offset per expert after padding (l,)
        alpha_tensor: Per-group alpha scaling
        beta_tensor: Per-group beta scaling
        prob_tensor: Per-row probability tensor
        norm_const_tensor: Optional normalization constant
        acc_dtype: Accumulator data type
        d_dtype: Output D tensor data type
        cd_major: CD major dimension (note: only "n"-major layout is supported)
        mma_tiler_mn: MMA tiler shape
        cluster_shape_mn: Cluster shape
        sf_vec_size: Scale factor vector size
        vector_f32: Use vectorized f32
        m_aligned: M alignment
        discrete_col_sfd: Boolean, True to generate discrete col-major scale factor tensor
        epilogue_op: Optional epilogue operation. Valid values: None, "none", "identity", "relu", "srelu"
        current_stream: CUDA stream

    Returns:
        TupleDict: A dictionary-like object containing output tensors that can also be unpacked as a tuple.
            Dictionary keys (also the unpacking order):
            - **d_row_tensor** (torch.Tensor): Final output tensor after dSwiGLU
            - **d_col_tensor** (torch.Tensor): Column-wise output tensor
            - **dprob_tensor** (torch.Tensor): Gradient tensor for prob (shape `(valid_m, 1, 1)`)
            - **amax_tensor** (torch.Tensor or None): Absolute maximum values (shape l, 2, 1)
            - **sfd_row_tensor** (torch.Tensor or None): Row-wise scale factors for D
            - **sfd_col_tensor** (torch.Tensor or None): Column-wise scale factors for D
    """
    valid_m = a_tensor.shape[0]
    n, _, l = b_tensor.shape

    if cd_major != "n":
        raise ValueError(f"cd_major must be 'n', got {cd_major}")

    use_full_dynamic = os.environ.get("CUDNN_FE_GROUPED_GEMM_DYNAMIC_MNKL", "1") != "0"

    def stride_order(tensor: torch.Tensor) -> Tuple[int, ...]:
        return tuple(i for i, s in sorted(enumerate(tensor.stride()), key=lambda x: x[1]))

    cache_key = (
        use_full_dynamic,
        a_tensor.shape[1:] if not use_full_dynamic else None,
        b_tensor.shape if not use_full_dynamic else None,
        c_tensor.shape[1:] if not use_full_dynamic else None,
        a_tensor.dtype,
        b_tensor.dtype,
        c_tensor.dtype,
        stride_order(a_tensor),
        stride_order(b_tensor),
        stride_order(c_tensor),
        sfa_tensor.dtype,
        sfb_tensor.dtype,
        padded_offsets.shape if not use_full_dynamic else None,
        padded_offsets.stride() if not use_full_dynamic else None,
        padded_offsets.dtype,
        norm_const_tensor.shape if norm_const_tensor is not None else None,
        norm_const_tensor.stride() if norm_const_tensor is not None else None,
        norm_const_tensor.dtype if norm_const_tensor is not None else None,
        acc_dtype,
        d_dtype,
        cd_major,
        mma_tiler_mn,
        cluster_shape_mn,
        sf_vec_size,
        vector_f32,
        m_aligned,
        discrete_col_sfd,
        epilogue_op,
    )

    # Allocate M-dependent output tensors fresh every call (M varies across MoE steps).
    # Only M-independent tensors (amax, beta) are cached to avoid repeated allocation.
    _logger.debug("grouped_gemm_dswiglu_wrapper_sm100: Allocating M-dependent output tensors")
    d_row_tensor = torch.empty_strided((valid_m, n * 2, 1), (n * 2, 1, valid_m * n * 2), dtype=d_dtype, device=a_tensor.device)
    d_col_tensor = torch.empty_strided((valid_m, n * 2, 1), (n * 2, 1, valid_m * n * 2), dtype=d_dtype, device=a_tensor.device)
    dprob_tensor = dprob_tensor_buf.zero_() if dprob_tensor_buf is not None else torch.zeros((valid_m, 1, 1), dtype=torch.float32, device=a_tensor.device)

    if valid_m == 0:
        amax_tensor = None
        if d_dtype in [torch.bfloat16, torch.float16]:
            amax_tensor = torch.full((l, 2, 1), float("-inf"), dtype=torch.float32, device=a_tensor.device)
        _logger.debug("grouped_gemm_dswiglu_wrapper_sm100: valid_m is zero, skipping kernel execution")
        return TupleDict(
            d_row_tensor=d_row_tensor,
            d_col_tensor=d_col_tensor,
            dprob_tensor=dprob_tensor,
            amax_tensor=amax_tensor,
            sfd_row_tensor=None,
            sfd_col_tensor=None,
        )

    sfd_row_tensor = None
    sfd_col_tensor = None
    if a_tensor.dtype in [torch.float8_e4m3fn, torch.float8_e5m2] and sfa_tensor.dtype in [torch.float8_e8m0fnu, torch.float8_e4m3fn]:
        _logger.debug("grouped_gemm_dswiglu_wrapper_sm100: Detected fp8 a_dtype and sfa_dtype, constructing sfd_row_tensor and sfd_col_tensor")
        sf_dtype = sfa_tensor.dtype
        mma_permute_order = (3, 4, 1, 5, 2, 0)
        sf_k_row = ceil_div(n * 2, sf_vec_size)
        mma_shape_row = (1, ceil_div(valid_m, 128), ceil_div(sf_k_row, 4), 32, 4, 4)
        sfd_row_tensor = torch.empty(mma_shape_row, dtype=sf_dtype, device=a_tensor.device).permute(mma_permute_order)
        sf_k_col = ceil_div(valid_m, sf_vec_size)
        mma_shape_col = (1, ceil_div(n * 2, 128), ceil_div(sf_k_col, 4), 32, 4, 4)
        sfd_col_tensor = torch.empty(mma_shape_col, dtype=sf_dtype, device=a_tensor.device).permute(mma_permute_order)

    if cache_key in _cache_of_GroupedGemmDswigluSm100Objects:
        _logger.debug("group_gemm_dswiglu_wrapper_sm100: Using previously cached GroupedGemmDswigluSm100 object")
        grouped_gemm_dswiglu, cached_amax_tensor, cached_beta_tensor = _cache_of_GroupedGemmDswigluSm100Objects[cache_key]
        amax_tensor = amax_tensor_buf if amax_tensor_buf is not None else cached_amax_tensor
        if beta_tensor is not None:
            effective_beta = beta_tensor
        elif cached_beta_tensor is not None:
            effective_beta = cached_beta_tensor
        else:
            # Fallback: cache was populated without beta caching (non-NVFP4 path),
            # but caller now passes None (NVFP4 path). Create ones tensor on-the-fly.
            effective_beta = torch.ones(l, dtype=torch.float32, device=a_tensor.device)
    else:
        _logger.debug(
            "group_gemm_dswiglu_wrapper_sm100: No previously cached GroupedGemmDswigluSm100 object found, creating new GroupedGemmDswigluSm100 object"
        )

        # For NVFP4 (beta_tensor=None): create and cache a ones tensor — avoids FillFunctor on every step.
        # For non-NVFP4 (beta_tensor provided): use caller's value directly; don't cache (it changes each step).
        cached_beta_tensor = torch.ones(l, dtype=torch.float32, device=a_tensor.device) if beta_tensor is None else None
        effective_beta = cached_beta_tensor if beta_tensor is None else beta_tensor

        cached_amax_tensor = None
        amax_tensor = None
        if d_dtype in [torch.bfloat16, torch.float16]:
            _logger.debug("grouped_gemm_dswiglu_wrapper_sm100: Detected bf16/float16 d_dtype, constructing amax_tensor")
            cached_amax_tensor = torch.empty((l, 2, 1), dtype=torch.float32, device=a_tensor.device)
            amax_tensor = amax_tensor_buf if amax_tensor_buf is not None else cached_amax_tensor

        grouped_gemm_dswiglu = GroupedGemmDswigluSm100(
            sample_a=a_tensor,
            sample_b=b_tensor,
            sample_c=c_tensor,
            sample_d_row=d_row_tensor,
            sample_d_col=d_col_tensor,
            sample_sfa=sfa_tensor,
            sample_sfb=sfb_tensor,
            sample_padded_offsets=padded_offsets,
            sample_alpha=alpha_tensor,
            sample_beta=effective_beta,
            sample_prob=prob_tensor,
            sample_dprob=dprob_tensor,
            sample_amax=cached_amax_tensor,
            sample_sfd_row=sfd_row_tensor,
            sample_sfd_col=sfd_col_tensor,
            sample_norm_const=norm_const_tensor,
            acc_dtype=acc_dtype,
            mma_tiler_mn=mma_tiler_mn,
            cluster_shape_mn=cluster_shape_mn,
            sf_vec_size=sf_vec_size,
            vector_f32=vector_f32,
            m_aligned=m_aligned,
            discrete_col_sfd=discrete_col_sfd,
            epilogue_op=epilogue_op,
        )

        assert grouped_gemm_dswiglu.check_support(), "Unsupported configuration"
        grouped_gemm_dswiglu.compile()
        _cache_of_GroupedGemmDswigluSm100Objects[cache_key] = (grouped_gemm_dswiglu, cached_amax_tensor, cached_beta_tensor)

    grouped_gemm_dswiglu.execute(
        a_tensor=a_tensor,
        b_tensor=b_tensor,
        c_tensor=c_tensor,
        d_row_tensor=d_row_tensor,
        d_col_tensor=d_col_tensor,
        sfa_tensor=sfa_tensor,
        sfb_tensor=sfb_tensor,
        padded_offsets=padded_offsets,
        alpha_tensor=alpha_tensor,
        beta_tensor=effective_beta,
        prob_tensor=prob_tensor,
        dprob_tensor=dprob_tensor,
        sfd_row_tensor=sfd_row_tensor,
        sfd_col_tensor=sfd_col_tensor,
        amax_tensor=amax_tensor,
        norm_const_tensor=norm_const_tensor,
        current_stream=current_stream,
    )

    return TupleDict(
        d_row_tensor=d_row_tensor,
        d_col_tensor=d_col_tensor,
        dprob_tensor=dprob_tensor,
        amax_tensor=amax_tensor,
        sfd_row_tensor=sfd_row_tensor,
        sfd_col_tensor=sfd_col_tensor,
    )



"""
Sparse Score Recompute Kernel — SM100 Cute-DSL Implementation.

Dual-mode kernel controlled by score_type ("indexer" or "attention"):

  Indexer mode (score_type="indexer"):
    S[b,q,i] = sum_h [ReLU(Q_h · K_{topk[b,q,i]}^T) · W_{b,q,h}]
    predict[b,q,:] = softmax(S[b,q,:])  (over topk dim)

  Attention mode (score_type="attention"):
    P[b,q,h,i] = exp(Q_h · K_{topk[b,q,i]}^T · scale - LSE[b,q,h])
    S[b,q,i] = sum_h P[b,q,h,i]
    target[b,q,:] = S[b,q,:] / sum(S[b,q,:])  (L1-norm over topk dim)

K is sparse-loaded via cp.async based on topk_indices (not TMA).

Design: SwapAB (K as A, Q_packed as B), PackGQA,
        1 epilogue warpgroup.
        WG2 (128 threads) loads K in SM90-style gather (16 groups x 8 threads).
        CLC persistent scheduling.
        Optional k_block_size splits head_dim into chunks for reduced sK SMEM.
"""

import math
from functools import partial

import cuda.bindings.driver as cuda

import cutlass
import cutlass.cute as cute
from cutlass import Float32, Int32, const_expr
from cutlass.cute.nvgpu import cpasync
import cutlass.cute.nvgpu.tcgen05 as tcgen05
from cutlass.pipeline import (
    Agent,
    CooperativeGroup,
    PipelineTmaUmma,
    PipelineAsyncUmma,
    PipelineUserType,
    make_pipeline_state,
    pipeline_init_arrive,
    pipeline_init_wait,
    PipelineClcFetchAsync,
)
import cutlass.utils as utils
from cutlass.utils.blackwell_helpers import (
    make_trivial_tiled_mma as _make_trivial_tiled_mma,
    make_smem_layout_a as _make_smem_layout_a,
    make_smem_layout_b as _make_smem_layout_b,
)

from cudnn.deepseek_sparse_attention.utils.sm100.gemm import gemm_ptx_partial as _gemm_ptx_partial
from cudnn.deepseek_sparse_attention.utils import copy as copy_ops

mul_packed_f32x2 = partial(cute.arch.mul_packed_f32x2, rnd="rn")
add_packed_f32x2 = partial(cute.arch.add_packed_f32x2, rnd="rn")
fma_packed_f32x2 = partial(cute.arch.fma_packed_f32x2, rnd="rn")


class SparseScoreRecomputeSm100:
    """
    SM100 Cute-DSL kernel for sparse backward score recomputation.

    Dual-mode via score_type:
      - "indexer": ReLU(QK) * W -> head reduce -> softmax
      - "attention": exp(QK*scale - LSE) -> head reduce -> L1-norm

    SwapAB design:
      - A = K (n_tile x head_dim), sparse-loaded via cp.async
      - B = Q_packed (m_tile x head_dim), loaded via TMA
      - C = S^T in TMEM: (n_tile, m_tile)

    Warp layout (12 warps total):
      - Warp 0:     Load (TMA Q, per-head data, topk indices)
      - Warp 1:     MMA  (QK GEMM via TCGen05, swapAB)
      - Warp 2:     CLC scheduler (producer)
      - Warp 3:     Idle
      - Warps 4-7:  Epilogue warpgroup (score reduction + normalization)
      - Warps 8-11: K loading warpgroup (sparse cp.async gather)
    """

    arch = 100
    WARP_SIZE = 32
    WARPGROUP_SIZE = 128

    def __init__(
        self,
        head_dim: int,
        qhead_per_kvhead: int = 32,
        m_block_size: int = 32,
        n_block_size: int = 128,
        topk: int = 512,
        kv_stage: int = 4,
        score_type: str = "indexer",
        have_topk_length: bool = False,
        topk_in_smem: bool = True,
        k_block_size: int | None = None,
        topk_indices_global: bool = True,
    ):
        assert score_type in ("indexer", "attention")
        self.score_type = score_type
        self.head_dim = head_dim
        self.qhead_per_kvhead = qhead_per_kvhead
        self.m_block_size = m_block_size
        self.n_block_size = n_block_size
        self.topk = topk
        self.kv_stage = kv_stage
        self.topk_indices_global = topk_indices_global
        self.sched_warp_id = 2
        self.num_clc_stage = 1
        self.num_clc_response_bytes = 16
        assert topk % n_block_size == 0, f"topk ({topk}) must be a multiple of n_block_size ({n_block_size})"
        self.num_n_blocks = topk // n_block_size
        self.have_topk_length = have_topk_length
        self.topk_in_smem = topk_in_smem

        # Pad head_dim to multiple of 16
        hdim_multiple_of = 16
        self.head_dim_padded = int(math.ceil(head_dim / hdim_multiple_of) * hdim_multiple_of)
        assert self.head_dim_padded % 64 == 0, (
            f"head_dim_padded ({self.head_dim_padded}) must be a multiple of 64 " f"(K loading uses 8 threads × 8 elements = 64 per iteration)"
        )

        # K-dimension splitting: process head_dim in num_k_chunks chunks of
        # k_block_size each, reducing sK SMEM per stage and enabling more
        # kv_stages for better K load / MMA overlap.
        self.k_block_size = k_block_size if k_block_size is not None else self.head_dim_padded
        assert self.head_dim_padded % self.k_block_size == 0, (
            f"head_dim_padded ({self.head_dim_padded}) must be a multiple of " f"k_block_size ({self.k_block_size})"
        )
        assert self.k_block_size % 64 == 0, (
            f"k_block_size ({self.k_block_size}) must be a multiple of 64 " f"(K loading uses 8 threads × 8 elements = 64 per iteration)"
        )
        self.num_k_chunks = self.head_dim_padded // self.k_block_size

        self.q_tokens_per_tile = m_block_size // qhead_per_kvhead
        assert self.q_tokens_per_tile == 1, "Only 1 q_token per tile is supported"

        # swapAB: K is A (M-dim=n_block), Q_packed is B (N-dim=m_block)
        # K dimension of the MMA tiler is k_block_size (not full head_dim).
        self.mma_tiler_qk = (n_block_size, m_block_size, self.k_block_size)
        self.qk_acc_dtype = Float32

        # Warp config: 12 warps
        self.load_warp_id = 0
        self.mma_warp_id = 1
        self.epilogue_wg_warp_ids = (4, 5, 6, 7)
        self.k_load_wg_warp_ids = (8, 9, 10, 11)
        self.num_warps = 12
        self.threads_per_cta = self.WARP_SIZE * self.num_warps  # 384
        # Register allocation
        # NOTE: num_regs_epilogue=256 accommodates pre-loading all per-head
        # data (W/LSE) into registers before the n_block loop, which requires
        # m_block_size BF16 (indexer) or m_block_size F32 (attention) extra regs.
        # With m_block_size=qhead_per_kvhead=64, this needs 64 BF16=32 F32 regs
        # (indexer) or 64 F32 regs (attention), fitting within 256. For
        # qhead_per_kvhead=128, this would need 128 BF16=64 or 128 F32 regs,
        # likely causing spills. In that case, fall back to the per-n_block
        # autovec_copy pattern or use a chunked pre-load strategy.
        self.num_regs_load = 48
        self.num_regs_mma = 48
        self.num_regs_epilogue = 256

        # TMEM: accumulator slots reused across n_blocks (multi-round when
        # num_n_blocks > num_tmem_slots). S barriers are sized to match slots,
        # with phase cycling handling cross-round synchronization.
        SM100_TMEM_CAPACITY_COLUMNS = 512
        self.tmem_s_stride = self.m_block_size
        self.num_tmem_slots = min(
            self.num_n_blocks,
            SM100_TMEM_CAPACITY_COLUMNS // self.m_block_size,
        )
        self.tmem_total = self.tmem_s_stride * self.num_tmem_slots
        self.tmem_alloc_cols = SM100_TMEM_CAPACITY_COLUMNS

        # Pipeline barrier counts (Q uses 1 stage with combined tx_count)
        self.Q_mbar_size = 2
        self.K_mbar_size = 2 * self.kv_stage
        # S barriers: paired (s_full, s_empty) per TMEM slot (not per n_block).
        # Layout: [s_full[0], s_empty[0], s_full[1], s_empty[1], ...]
        # When num_n_blocks > num_tmem_slots, slots and barriers are reused
        # across rounds with phase cycling.
        self.S_mbar_size = 2 * self.num_tmem_slots
        self.s_empty_arrive_count = self.WARPGROUP_SIZE

        # Cross-warp reduce sync barrier: warpgroup (128 threads) sync during score reduction
        self.reduce_sync_mbar_size = 2
        self.reduce_sync_arrive_count = self.WARPGROUP_SIZE

        # sScoreAll: cross-warp scratch for reduce (one entry per warp in epilogue warpgroup)
        # Indexer epilogue uses 2 reduce passes (max + sum) that reuse the same buffer.
        # Allocate 2x to give each pass its own region, avoiding potential reuse races.
        self.num_warps_in_epi_wg = len(self.epilogue_wg_warp_ids)
        self.sScoreAll_size = self.num_warps_in_epi_wg * 2

        # SMEM alignment
        self.buffer_align_bytes = 1024
        # Cluster shape (no multi-CTA clustering for simplicity)
        self.cluster_shape_mn = (1, 1)

    # =========================================================================
    # Host-side entry: layout transpose, PackGQA, TMA creation, launch
    # =========================================================================
    @cute.jit
    def __call__(
        self,
        mQ: cute.Tensor,  # (bs, seqlen_q, n_heads_q, head_dim) BF16
        mK: cute.Tensor,  # (bs, seqlen_k, head_dim) BF16
        mPerHead: cute.Tensor,  # (bs, seqlen_q, n_heads_q) — W (BF16) or scaled_LSE (FP32)
        mTopkIdx: cute.Tensor,  # (bs, seqlen_q, topk) INT32
        mOut: cute.Tensor,  # (bs, seqlen_q, topk) FP32
        mTopkLength: cute.Tensor,  # (bs, seqlen_q) INT32 (dummy when unused)
        softmax_scale: Float32 | float,
        stream: cuda.CUstream,
    ):
        """Host-side: layout transpose, PackGQA, TMA creation, kernel launch."""

        self.q_dtype = mQ.element_type
        self.k_dtype = mK.element_type
        self.per_head_dtype = mPerHead.element_type

        self.tma_copy_bytes = {
            "Q": self.m_block_size * self.head_dim_padded * (self.q_dtype.width // 8),
        }

        # --- Layout transpose: (bs, seqlen, heads, hdim) -> (seqlen, hdim, heads, bs) ---
        mQ = cute.make_tensor(mQ.iterator, cute.select(mQ.layout, mode=[1, 3, 2, 0]))

        # --- PackGQA: reshape Q to pack qhpkv heads into seqlen_q dim ---
        shape_Q_packed = (
            (self.qhead_per_kvhead, mQ.shape[0]),  # packed M dim
            mQ.shape[1],  # head_dim
            1,  # n_heads_kv
            *mQ.shape[3:],  # bs (and beyond)
        )
        stride_Q_packed = (
            (mQ.stride[2], mQ.stride[0]),  # (stride_head, stride_seqlen)
            mQ.stride[1],  # stride_hdim
            mQ.stride[2] * self.qhead_per_kvhead,  # stride across kv_head groups
            *mQ.stride[3:],  # stride_bs
        )
        mQ = cute.make_tensor(mQ.iterator, cute.make_layout(shape_Q_packed, stride=stride_Q_packed))

        # --- K layout: (bs, seqlen_k, head_dim) -> (seqlen_k, head_dim, bs) ---
        mK = cute.make_tensor(mK.iterator, cute.select(mK.layout, mode=[1, 2, 0]))

        cta_group = tcgen05.CtaGroup.ONE
        self.q_major_mode = cutlass.utils.LayoutEnum.from_tensor(mQ).mma_major_mode()

        tiled_mma_qk = _make_trivial_tiled_mma(
            self.q_dtype,
            tcgen05.OperandMajorMode.K,
            self.q_major_mode,
            self.qk_acc_dtype,
            cta_group,
            self.mma_tiler_qk[:2],
        )

        self.cluster_shape_mnk = (*self.cluster_shape_mn, 1)
        self.cluster_layout_vmnk = cute.tiled_divide(
            cute.make_layout(self.cluster_shape_mnk),
            (tiled_mma_qk.thr_id.shape,),
        )

        # --- SMEM layouts --- (K is A operand in swapAB; loaded via cp.async, not TMA)
        sK_layout = _make_smem_layout_a(
            tiled_mma_qk,
            self.mma_tiler_qk,
            self.k_dtype,
            self.kv_stage,
        )
        sQ_layout = _make_smem_layout_b(
            tiled_mma_qk,
            self.mma_tiler_qk,
            self.q_dtype,
            self.num_k_chunks,
        )

        # TMA atom for Q only (K is sparse-loaded, no TMA)
        tma_load_op = cpasync.CopyBulkTensorTileG2SOp(cta_group)
        tma_atom_Q, mQ = cute.nvgpu.make_tiled_tma_atom_B(
            tma_load_op,
            mQ,
            cute.select(sQ_layout, mode=[0, 1, 2]),
            self.mma_tiler_qk,
            tiled_mma_qk,
            self.cluster_layout_vmnk.shape,
        )

        self.sQ_layout = sQ_layout
        self.sK_layout = sK_layout

        # --- PerHead layout: (bs, seqlen_q, n_heads_q) -> (seqlen_q, n_heads_q, bs) ---
        mPerHead = cute.make_tensor(mPerHead.iterator, cute.select(mPerHead.layout, mode=[1, 2, 0]))

        # --- TopkIdx layout: (bs, seqlen_q, topk) -> (seqlen_q, topk, bs) ---
        mTopkIdx = cute.make_tensor(mTopkIdx.iterator, cute.select(mTopkIdx.layout, mode=[1, 2, 0]))

        # --- Output layout: (bs, seqlen_q, topk) -> (seqlen_q, topk, bs) ---
        mOut = cute.make_tensor(mOut.iterator, cute.select(mOut.layout, mode=[1, 2, 0]))

        # --- TopkLength layout: (bs, seqlen_q) -> (seqlen_q, bs) ---
        mTopkLength = cute.make_tensor(mTopkLength.iterator, cute.select(mTopkLength.layout, mode=[1, 0]))

        # --- Grid and kernel dispatch (CLC persistent scheduling) ---
        seqlen_q_packed = cute.size(mQ.shape[0])
        num_m_blocks = cute.ceil_div(seqlen_q_packed, self.m_block_size)
        batch_size = cute.size(mQ.shape[3]) if cute.rank(mQ.shape) > 3 else 1
        tile_sched_params = utils.ClcDynamicPersistentTileSchedulerParams((num_m_blocks, 1, batch_size), (*self.cluster_shape_mn, 1))
        grid_dim = utils.ClcDynamicPersistentTileScheduler.get_grid_shape(tile_sched_params)
        self.kernel(
            mQ,
            mK,
            mPerHead,
            mTopkIdx,
            mOut,
            mTopkLength,
            softmax_scale,
            tma_atom_Q,
            tiled_mma_qk,
            sQ_layout,
            sK_layout,
            tile_sched_params,
        ).launch(
            grid=grid_dim,
            block=[self.threads_per_cta, 1, 1],
            cluster=self.cluster_shape_mnk,
            stream=stream,
        )

    # =========================================================================
    # Device-side kernel
    # =========================================================================
    @cute.kernel
    def kernel(
        self,
        mQ,
        mK,
        mPerHead,
        mTopkIdx,
        mOut,
        mTopkLength,
        softmax_scale: Float32 | float,
        tma_atom_Q,
        tiled_mma_qk,
        sQ_layout,
        sK_layout,
        tile_sched_params: utils.ClcDynamicPersistentTileSchedulerParams,
    ):
        """Device-side kernel entry with CLC persistent scheduling."""
        seqlen_q_packed = cute.size(mQ.shape[0])
        seqlen_k = cute.size(mK.shape[0])
        warp_idx = cute.arch.make_warp_uniform(cute.arch.warp_idx())
        tidx = cute.arch.thread_idx()[0]

        # --- TMA descriptor prefetch ---
        if warp_idx == 0:
            cpasync.prefetch_descriptor(tma_atom_Q)

        # =====================================================================
        # Shared memory allocation
        # =====================================================================
        sQ_size = cute.cosize(sQ_layout)
        sK_size = cute.cosize(sK_layout)
        sPerHead_size = self.m_block_size * 2  # double-buffer for LOAD/EPI overlap
        sTopkIdx_size = self.topk * 2 if self.topk_in_smem else 1

        @cute.struct
        class SharedStorage:
            Q_mbar_ptr: cute.struct.MemRange[cutlass.Int64, self.Q_mbar_size]
            K_mbar_ptr: cute.struct.MemRange[cutlass.Int64, self.K_mbar_size]
            S_mbar_ptr: cute.struct.MemRange[cutlass.Int64, self.S_mbar_size]
            reduce_sync_mbar_ptr: cute.struct.MemRange[cutlass.Int64, self.reduce_sync_mbar_size]
            tmem_dealloc_mbar_ptr: cute.struct.MemRange[cutlass.Int64, 1]
            tmem_holding_buf: Int32
            clc_mbar_ptr: cute.struct.MemRange[cutlass.Int64, 2]
            clc_response: cute.struct.MemRange[cutlass.Int32, 4]
            sPerHead: cute.struct.Align[
                cute.struct.MemRange[self.per_head_dtype, sPerHead_size],
                128,
            ]
            sTopkIdx: cute.struct.Align[
                cute.struct.MemRange[Int32, sTopkIdx_size],
                128,
            ]
            sScoreAll: cute.struct.Align[
                cute.struct.MemRange[Float32, self.sScoreAll_size],
                self.buffer_align_bytes,
            ]
            sQ: cute.struct.Align[
                cute.struct.MemRange[self.q_dtype, sQ_size],
                self.buffer_align_bytes,
            ]
            sK: cute.struct.Align[
                cute.struct.MemRange[self.k_dtype, sK_size],
                self.buffer_align_bytes,
            ]

        smem = cutlass.utils.SmemAllocator()
        storage = smem.allocate(SharedStorage)

        # Extract all pointers/tensors from storage so no 'if' body references storage
        # (avoids DSLRuntimeError: SharedStorage as user-defined Python object in if)
        Q_mbar_ptr = storage.Q_mbar_ptr.data_ptr()
        K_mbar_ptr = storage.K_mbar_ptr.data_ptr()
        S_mbar_ptr = storage.S_mbar_ptr.data_ptr()
        reduce_sync_mbar_ptr = storage.reduce_sync_mbar_ptr.data_ptr()
        tmem_dealloc_mbar_ptr = storage.tmem_dealloc_mbar_ptr.data_ptr()
        tmem_holding_buf = storage.tmem_holding_buf
        clc_mbar_ptr = storage.clc_mbar_ptr.data_ptr()
        clc_response_ptr = storage.clc_response.data_ptr()
        sK = storage.sK.get_tensor(sK_layout.outer, swizzle=sK_layout.inner)
        sQ = storage.sQ.get_tensor(sQ_layout.outer, swizzle=sQ_layout.inner)
        sPerHead = storage.sPerHead.get_tensor(cute.make_layout((self.m_block_size,), stride=(1,)))
        sTopkIdx = storage.sTopkIdx.get_tensor(cute.make_layout((self.topk if self.topk_in_smem else 1,), stride=(1,)))
        sScoreAll = storage.sScoreAll.get_tensor(cute.make_layout((self.sScoreAll_size,), stride=(1,)))

        # =====================================================================
        # Pipeline setup
        # =====================================================================
        cluster_layout_vmnk = cute.tiled_divide(
            cute.make_layout(self.cluster_shape_mnk),
            (tiled_mma_qk.thr_id.shape,),
        )
        cta_rank_in_cluster = cute.arch.make_warp_uniform(cute.arch.block_idx_in_cluster())
        is_first_cta_in_cluster = cta_rank_in_cluster == 0

        # Q pipeline (1 stage: all k_chunks share one barrier with combined tx_count)
        pipeline_Q = PipelineTmaUmma.create(
            barrier_storage=Q_mbar_ptr,
            num_stages=1,
            producer_group=CooperativeGroup(Agent.Thread, 1),
            consumer_group=CooperativeGroup(Agent.Thread, 1),
            tx_count=self.tma_copy_bytes["Q"],
            cta_layout_vmnk=cluster_layout_vmnk,
            defer_sync=True,
        )

        # K pipeline (cp.async, WG2 = 128 threads)
        pipeline_K = PipelineAsyncUmma.create(
            barrier_storage=K_mbar_ptr,
            num_stages=self.kv_stage,
            producer_group=CooperativeGroup(Agent.Thread, self.WARPGROUP_SIZE),
            consumer_group=CooperativeGroup(Agent.Thread, 1),
            defer_sync=True,
        )

        # S barriers: paired (s_full, s_empty) per TMEM slot
        if warp_idx == 1:
            cute.arch.mbarrier_init(tmem_dealloc_mbar_ptr, 1)
        if warp_idx == 0:
            for _si in cutlass.range_constexpr(self.num_tmem_slots):
                cute.arch.mbarrier_init(S_mbar_ptr + 2 * _si, 1)
                cute.arch.mbarrier_init(S_mbar_ptr + 2 * _si + 1, self.s_empty_arrive_count)
            cute.arch.mbarrier_init(reduce_sync_mbar_ptr, self.reduce_sync_arrive_count)

        # CLC persistent scheduling pipeline
        cluster_size = cute.size(self.cluster_shape_mn)
        num_clc_consumer_threads = self.WARP_SIZE * (1 + cluster_size * (1 + 1 + 8))  # sched(1) + load(1) + mma(1) + epilogue+Kload(8) warps
        clc_pipeline = PipelineClcFetchAsync.create(
            barrier_storage=clc_mbar_ptr,
            num_stages=self.num_clc_stage,
            producer_group=CooperativeGroup(Agent.Thread),
            consumer_group=CooperativeGroup(Agent.Thread, num_clc_consumer_threads),
            tx_count=self.num_clc_response_bytes,
            cta_layout_vmnk=cluster_layout_vmnk,
            defer_sync=True,
        )

        pipeline_init_arrive(cluster_shape_mn=cluster_layout_vmnk, is_relaxed=True)
        cute.arch.sync_threads()
        pipeline_init_wait(cluster_shape_mn=cluster_layout_vmnk)

        clc_consumer_state = make_pipeline_state(PipelineUserType.Consumer, self.num_clc_stage)
        tile_sched = utils.ClcDynamicPersistentTileScheduler.create(
            tile_sched_params,
            cute.arch.block_idx(),
            cute.arch.grid_dim(),
            clc_response_ptr,
        )
        work_tile = tile_sched.initial_work_tile_info()

        Q_producer, Q_consumer = pipeline_Q.make_participants()

        # TMEM accumulator layout reference (per-n_block tensors constructed in epilogue)
        thr_mma_qk = tiled_mma_qk.get_slice(0)
        qk_acc_shape = thr_mma_qk.partition_shape_C(self.mma_tiler_qk[:2])
        tStS_fake = thr_mma_qk.make_fragment_C(qk_acc_shape)
        tmem_ptr = cute.make_ptr(Float32, 0, mem_space=cute.AddressSpace.tmem, assumed_align=16)
        tStS_ref = cute.make_tensor(tmem_ptr, tStS_fake.layout)

        warp_group_idx = tidx // self.WARPGROUP_SIZE

        # =====================================================================
        # Warpgroup 0 (warps 0-3): Load + MMA + Scheduler
        # =====================================================================
        if warp_group_idx == 0:
            cute.arch.setmaxregister_decrease(self.num_regs_load)

            # -----------------------------------------------------------------
            # Load warp (warp 0): PerHead data + TopkIdx + TMA Q
            # -----------------------------------------------------------------
            if warp_idx == self.load_warp_id:
                seqlen_q_load = seqlen_q_packed // self.qhead_per_kvhead
                rows_per_thread = cute.ceil_div(self.m_block_size, self.WARP_SIZE)
                topk_rows_per_thread = cute.ceil_div(self.topk, self.WARP_SIZE)
                lane_id = tidx % self.WARP_SIZE
                tile_count = Int32(0)

                while work_tile.is_valid_tile:
                    m_block = work_tile.tile_idx[0]
                    batch_idx = work_tile.tile_idx[2]

                    # PerHead data load (double-buffered)
                    per_head_buf_off = (tile_count % 2) * self.m_block_size
                    for ri in cutlass.range_constexpr(rows_per_thread):
                        row = ri * self.WARP_SIZE + lane_id
                        if row < self.m_block_size:
                            m_packed_idx = m_block * self.m_block_size + row
                            m_idx = m_packed_idx // self.qhead_per_kvhead
                            h_idx = m_packed_idx - m_idx * self.qhead_per_kvhead
                            if m_idx < seqlen_q_load:
                                sPerHead[per_head_buf_off + row] = mPerHead[m_idx, h_idx, batch_idx]
                            else:
                                sPerHead[per_head_buf_off + row] = self.per_head_dtype(0)

                    if const_expr(self.topk_in_smem):
                        # TopkIdx load (double-buffered)
                        topk_idx_buf_off = (tile_count % 2) * self.topk
                        for ri in cutlass.range(topk_rows_per_thread, unroll_full=True):
                            topk_pos = ri * self.WARP_SIZE + lane_id
                            if topk_pos < self.topk:
                                if m_block < seqlen_q_load:
                                    sTopkIdx[topk_idx_buf_off + topk_pos] = mTopkIdx[m_block, topk_pos, batch_idx]
                                else:
                                    sTopkIdx[topk_idx_buf_off + topk_pos] = Int32(-1)
                    cute.arch.fence_view_async_shared()

                    # TMA Q load (1 barrier, num_k_chunks TMA copies)
                    Q_producer.reset()
                    mQ_cur = mQ[None, None, 0, batch_idx]
                    handle_Q = Q_producer.acquire_and_advance()
                    for _kc in cutlass.range_constexpr(self.num_k_chunks):
                        gQ = cute.local_tile(
                            mQ_cur,
                            (self.m_block_size, self.k_block_size),
                            (m_block, _kc),
                        )
                        sQ_cur = sQ[None, None, None, _kc]
                        load_Q_fn, _, _ = copy_ops.tma_get_copy_fn(
                            tma_atom_Q,
                            0,
                            cute.make_layout(1),
                            gQ,
                            sQ_cur,
                            single_stage=True,
                        )
                        load_Q_fn(tma_bar_ptr=handle_Q.barrier)

                    clc_pipeline.consumer_wait(clc_consumer_state)
                    work_tile = tile_sched.get_current_work()
                    clc_pipeline.consumer_release(clc_consumer_state)
                    tile_count = tile_count + 1
                    clc_consumer_state.advance()
                Q_producer.tail()

            # -----------------------------------------------------------------
            # MMA warp (warp 1): QK GEMM via TCGen05
            # -----------------------------------------------------------------
            if warp_idx == self.mma_warp_id:
                tmem_alloc_cols = Int32(self.tmem_alloc_cols)
                cute.arch.alloc_tmem(tmem_alloc_cols, tmem_holding_buf)
                cute.arch.sync_warp()

                s_empty_phase = Int32(1)  # unblock first round's s_empty barriers
                NUM_SLOTS_MASK = Int32(self.num_tmem_slots - 1)
                K_mma_state = make_pipeline_state(PipelineUserType.Consumer, self.kv_stage)
                while work_tile.is_valid_tile:
                    m_block = work_tile.tile_idx[0]
                    batch_idx = work_tile.tile_idx[2]

                    # Wait for all Q k_chunks (single barrier covers all TMA copies)
                    Q_consumer.reset()
                    handle_Q = Q_consumer.wait_and_advance()

                    tSrK = tiled_mma_qk.make_fragment_A(sK)
                    tSrQ = tiled_mma_qk.make_fragment_B(sQ)
                    qk_mma_op = tiled_mma_qk.op

                    q_token_idx = m_block
                    n_block = Int32(self.num_n_blocks - 1)

                    # Attention (hd=512): skip invalid blocks to save expensive MMA.
                    # Indexer (hd=128): process all blocks — static bounds enable better codegen.
                    if const_expr(self.have_topk_length and self.score_type == "attention"):
                        topK_mma = mTopkLength[q_token_idx, batch_idx]
                        n_block_max = (topK_mma + self.n_block_size - 1) // self.n_block_size
                        while n_block >= n_block_max:
                            slot = n_block & NUM_SLOTS_MASK
                            cute.arch.mbarrier_wait(S_mbar_ptr + 2 * slot + 1, s_empty_phase)
                            with cute.arch.elect_one():
                                cute.arch.mbarrier_arrive(S_mbar_ptr + 2 * slot)
                            if slot == Int32(0):
                                s_empty_phase ^= 1
                            n_block = n_block - 1

                    while n_block >= Int32(0):
                        slot = n_block & NUM_SLOTS_MASK
                        cute.arch.mbarrier_wait(S_mbar_ptr + 2 * slot + 1, s_empty_phase)

                        for _kc in cutlass.range_constexpr(self.num_k_chunks):
                            pipeline_K.consumer_wait(K_mma_state)

                            tSrKi = tSrK[None, None, None, K_mma_state.index]
                            sK_cur = sK[None, None, None, K_mma_state.index]
                            tSrQ_kc = tSrQ[None, None, None, _kc]
                            sQ_kc = sQ[None, None, None, _kc]
                            _gemm_ptx_partial(
                                qk_mma_op,
                                slot * self.tmem_s_stride,
                                tSrKi,
                                tSrQ_kc,
                                sA=sK_cur,
                                sB=sQ_kc,
                                zero_init=const_expr(_kc == 0),
                            )

                            pipeline_K.consumer_release(K_mma_state)
                            K_mma_state.advance()

                        with cute.arch.elect_one():
                            tcgen05.commit(S_mbar_ptr + 2 * slot)

                        if slot == Int32(0):
                            s_empty_phase ^= 1
                        n_block = n_block - 1

                    handle_Q.release()
                    clc_pipeline.consumer_wait(clc_consumer_state)
                    work_tile = tile_sched.get_current_work()
                    clc_pipeline.consumer_release(clc_consumer_state)
                    clc_consumer_state.advance()

                # TMEM dealloc
                cute.arch.relinquish_tmem_alloc_permit()
                cute.arch.mbarrier_wait(tmem_dealloc_mbar_ptr, 0)
                tmem_ptr_dealloc = cute.arch.retrieve_tmem_ptr(
                    Float32,
                    alignment=16,
                    ptr_to_buffer_holding_addr=tmem_holding_buf,
                )
                cute.arch.dealloc_tmem(tmem_ptr_dealloc, Int32(self.tmem_alloc_cols))

            # -----------------------------------------------------------------
            # Scheduler warp (warp 2): CLC producer loop
            # -----------------------------------------------------------------
            if warp_idx == self.sched_warp_id and is_first_cta_in_cluster:
                clc_producer_state = make_pipeline_state(PipelineUserType.ProducerConsumer, self.num_clc_stage)
                while work_tile.is_valid_tile:
                    clc_pipeline.producer_acquire(clc_producer_state)
                    mbarrier_addr = clc_pipeline.producer_get_barrier(clc_producer_state)
                    tile_sched.advance_to_next_work(mbarrier_addr)
                    clc_producer_state.advance()
                    clc_pipeline.consumer_wait(clc_consumer_state)
                    work_tile = tile_sched.get_current_work()
                    clc_pipeline.consumer_release(clc_consumer_state)
                    clc_consumer_state.advance()
                clc_pipeline.producer_tail(clc_producer_state)

        # =====================================================================
        # Warpgroup 1 (warps 4-7): Epilogue — score reduction + normalization
        # =====================================================================
        if warp_group_idx == 1:
            cute.arch.setmaxregister_increase(self.num_regs_epilogue)
            s_full_phase = Int32(0)
            reduce_phase = Int32(0)
            tile_count = Int32(0)
            while work_tile.is_valid_tile:
                m_block = work_tile.tile_idx[0]
                batch_idx = work_tile.tile_idx[2]
                per_head_offset = (tile_count % 2) * self.m_block_size
                topk_idx_offset = (tile_count % 2) * self.topk
                if cutlass.const_expr(self.score_type == "attention"):
                    if cutlass.const_expr(self.n_block_size >= 128):
                        s_full_phase, reduce_phase = self._epilogue_attention_n128(
                            tiled_mma_qk,
                            tStS_ref,
                            sPerHead,
                            sScoreAll,
                            S_mbar_ptr,
                            reduce_sync_mbar_ptr,
                            sTopkIdx,
                            mTopkIdx,
                            mOut,
                            mTopkLength,
                            m_block,
                            batch_idx,
                            tidx,
                            s_full_phase,
                            reduce_phase,
                            softmax_scale,
                            per_head_offset=per_head_offset,
                            topk_idx_offset=topk_idx_offset,
                        )
                    else:
                        s_full_phase, reduce_phase = self._epilogue_attention(
                            tiled_mma_qk,
                            tStS_ref,
                            sPerHead,
                            sScoreAll,
                            S_mbar_ptr,
                            reduce_sync_mbar_ptr,
                            sTopkIdx,
                            mTopkIdx,
                            mOut,
                            mTopkLength,
                            m_block,
                            batch_idx,
                            tidx,
                            s_full_phase,
                            reduce_phase,
                            softmax_scale,
                            per_head_offset=per_head_offset,
                            topk_idx_offset=topk_idx_offset,
                        )
                else:
                    s_full_phase, reduce_phase = self._epilogue_indexer(
                        tiled_mma_qk,
                        tStS_ref,
                        mPerHead,
                        sPerHead,
                        sScoreAll,
                        S_mbar_ptr,
                        reduce_sync_mbar_ptr,
                        sTopkIdx,
                        mTopkIdx,
                        mOut,
                        mTopkLength,
                        m_block,
                        batch_idx,
                        tidx,
                        s_full_phase,
                        reduce_phase,
                        softmax_scale,
                        per_head_offset=per_head_offset,
                        topk_idx_offset=topk_idx_offset,
                    )
                clc_pipeline.consumer_wait(clc_consumer_state)
                work_tile = tile_sched.get_current_work()
                clc_pipeline.consumer_release(clc_consumer_state)
                tile_count = tile_count + 1
                clc_consumer_state.advance()
            if warp_idx == self.epilogue_wg_warp_ids[-1]:
                with cute.arch.elect_one():
                    cute.arch.mbarrier_arrive(tmem_dealloc_mbar_ptr)

        # =====================================================================
        # Warpgroup 2 (warps 8-11): Sparse K loading via cp.async
        # =====================================================================
        if warp_group_idx == 2:
            cute.arch.setmaxregister_decrease(self.num_regs_load)
            wg_tidx = tidx % self.WARPGROUP_SIZE

            async_copy_atom = cute.make_copy_atom(
                cpasync.CopyG2SOp(cache_mode=cpasync.LoadCacheMode.GLOBAL),
                self.k_dtype,
                num_bits_per_copy=128,
            )
            async_thr_copy = cute.make_tiled_copy_tv(
                async_copy_atom,
                cute.make_layout((1,)),
                cute.make_layout((8,)),
            ).get_slice(0)

            # SM90-style gather: 16 groups x 8 threads, each group loads ROWS_PER_GROUP rows
            GROUP_SIZE = const_expr(8)
            NUM_GROUPS = const_expr(self.WARPGROUP_SIZE // 8)
            ROWS_PER_GROUP = const_expr(self.n_block_size // NUM_GROUPS)
            idx_in_group = wg_tidx % GROUP_SIZE
            group_idx_local = wg_tidx // GROUP_SIZE

            if const_expr(self.score_type == "indexer"):
                # Indexer is K-load-bound (MMA 16 cy vs K gather ~2000 cy/stage).
                # 2-deep cp.async pipeline: issue stage s+1 before waiting for
                # stage s, keeping 2 groups in flight to improve L2 utilization.
                # (Full-depth tested: worse — delaying MMA start hurts more
                # than extra L2 parallelism helps.)
                K_issue_state = make_pipeline_state(PipelineUserType.Producer, self.kv_stage)
                K_commit_state = make_pipeline_state(PipelineUserType.Producer, self.kv_stage)

                while work_tile.is_valid_tile:
                    m_block = work_tile.tile_idx[0]
                    batch_idx = work_tile.tile_idx[2]
                    q_token_idx = m_block
                    topK_cur = Int32(self.topk)
                    n_block_max = (topK_cur + self.n_block_size - 1) // self.n_block_size

                    # -- Prologue: issue first n_block, no wait --
                    n_block = n_block_max - 1
                    for _kc in cutlass.range_constexpr(self.num_k_chunks):
                        pipeline_K.producer_acquire(K_issue_state)
                        sK_stage = sK[None, None, None, K_issue_state.index]
                        sK_slice = cute.composition(
                            sK_stage,
                            cute.make_layout((self.n_block_size, self.k_block_size)),
                        )
                        k_offset = const_expr(_kc * self.k_block_size)
                        self._load_k_rows(
                            mK,
                            sK_slice,
                            mTopkIdx,
                            mTopkLength,
                            async_copy_atom,
                            async_thr_copy,
                            n_block,
                            q_token_idx,
                            batch_idx,
                            seqlen_k,
                            topK_cur,
                            idx_in_group,
                            group_idx_local,
                            k_offset=k_offset,
                        )
                        cute.arch.cp_async_commit_group()
                        K_issue_state.advance()
                    n_block = n_block - 1

                    # -- Main: issue next, wait/commit previous --
                    while n_block >= Int32(0):
                        for _kc in cutlass.range_constexpr(self.num_k_chunks):
                            pipeline_K.producer_acquire(K_issue_state)
                            sK_stage = sK[None, None, None, K_issue_state.index]
                            sK_slice = cute.composition(
                                sK_stage,
                                cute.make_layout((self.n_block_size, self.k_block_size)),
                            )
                            k_offset = const_expr(_kc * self.k_block_size)
                            self._load_k_rows(
                                mK,
                                sK_slice,
                                mTopkIdx,
                                mTopkLength,
                                async_copy_atom,
                                async_thr_copy,
                                n_block,
                                q_token_idx,
                                batch_idx,
                                seqlen_k,
                                topK_cur,
                                idx_in_group,
                                group_idx_local,
                                k_offset=k_offset,
                            )
                            cute.arch.cp_async_commit_group()
                            cute.arch.cp_async_wait_group(1)
                            cute.arch.fence_view_async_shared()
                            pipeline_K.producer_commit(K_commit_state)
                            K_commit_state.advance()
                            K_issue_state.advance()
                        n_block = n_block - 1

                    # -- Drain: commit last issued stage --
                    cute.arch.cp_async_wait_group(0)
                    cute.arch.fence_view_async_shared()
                    pipeline_K.producer_commit(K_commit_state)
                    K_commit_state.advance()

                    clc_pipeline.consumer_wait(clc_consumer_state)
                    work_tile = tile_sched.get_current_work()
                    clc_pipeline.consumer_release(clc_consumer_state)
                    clc_consumer_state.advance()

            else:
                # Attention: original K load with wait_group(0).
                # 2-deep tested but regressed -14% (kv_stage=2 too tight,
                # code bloat from prologue/drain outweighs L2 overlap gain).
                K_load_state = make_pipeline_state(PipelineUserType.Producer, self.kv_stage)

                while work_tile.is_valid_tile:
                    m_block = work_tile.tile_idx[0]
                    batch_idx = work_tile.tile_idx[2]
                    q_token_idx = m_block

                    if const_expr(self.have_topk_length):
                        topK_cur = mTopkLength[q_token_idx, batch_idx]
                    else:
                        topK_cur = Int32(self.topk)
                    n_block_max = (topK_cur + self.n_block_size - 1) // self.n_block_size

                    n_block = n_block_max - 1
                    while n_block >= Int32(0):
                        for _kc in cutlass.range_constexpr(self.num_k_chunks):
                            pipeline_K.producer_acquire(K_load_state)
                            sK_stage = sK[None, None, None, K_load_state.index]
                            sK_slice = cute.composition(
                                sK_stage,
                                cute.make_layout((self.n_block_size, self.k_block_size)),
                            )
                            k_offset = const_expr(_kc * self.k_block_size)
                            self._load_k_rows(
                                mK,
                                sK_slice,
                                mTopkIdx,
                                mTopkLength,
                                async_copy_atom,
                                async_thr_copy,
                                n_block,
                                q_token_idx,
                                batch_idx,
                                seqlen_k,
                                topK_cur,
                                idx_in_group,
                                group_idx_local,
                                k_offset=k_offset,
                            )
                            cute.arch.cp_async_commit_group()
                            cute.arch.cp_async_wait_group(0)
                            cute.arch.fence_view_async_shared()
                            pipeline_K.producer_commit(K_load_state)
                            K_load_state.advance()
                        n_block = n_block - 1

                    clc_pipeline.consumer_wait(clc_consumer_state)
                    work_tile = tile_sched.get_current_work()
                    clc_pipeline.consumer_release(clc_consumer_state)
                    clc_consumer_state.advance()

    # =========================================================================
    # Sparse K loading helpers (16 groups x 8 threads per row)
    # =========================================================================
    @cute.jit
    def _load_k_rows(
        self,
        mK: cute.Tensor,
        sK_slice: cute.Tensor,
        mTopkIdx: cute.Tensor,
        mTopkLength: cute.Tensor,
        copy_atom: cute.CopyAtom,
        thr_copy: cute.TiledCopy,
        n_block: Int32,
        q_token_idx: Int32,
        batch_idx: Int32,
        seqlen_k: Int32,
        topK_cur: Int32,
        idx_in_group: Int32,
        group_idx_local: Int32,
        k_offset: int = 0,
    ):
        """Load all K rows for one pipeline stage (one n_block × one k_chunk).

        When ``self.topk_indices_global`` (default), ``mTopkIdx`` carries the
        public global KV ids (``b * seqlen_k + local``); we decode back to
        local-per-batch before indexing into mK ``(seqlen_k, head_dim, bs)``.
        With the flag False, ids are already local (legacy callers — e.g.
        ``indexer_backward``'s grad path that still expects local).
        """
        NUM_GROUPS = const_expr(self.WARPGROUP_SIZE // 8)
        ROWS_PER_GROUP = const_expr(self.n_block_size // NUM_GROUPS)
        batch_offset = batch_idx * seqlen_k if const_expr(self.topk_indices_global) else Int32(0)
        for r in cutlass.range_constexpr(ROWS_PER_GROUP):
            row = r * NUM_GROUPS + group_idx_local
            topk_pos = n_block * self.n_block_size + row
            if const_expr(self.have_topk_length and self.score_type == "attention"):
                if topk_pos < topK_cur:
                    topk_raw = Int32(mTopkIdx[q_token_idx, topk_pos, batch_idx])
                    topk_local = topk_raw - batch_offset
                    self._copy_row(
                        mK,
                        sK_slice,
                        row,
                        idx_in_group,
                        copy_atom,
                        thr_copy,
                        topk_local,
                        batch_idx,
                        k_offset=k_offset,
                    )
            else:
                topk_local = Int32(-1)
                if topk_pos < topK_cur:
                    topk_raw = Int32(mTopkIdx[q_token_idx, topk_pos, batch_idx])
                    # Invalid (-1) stays negative after the offset subtraction
                    # and is rejected by the bounds check below.
                    topk_local = topk_raw - batch_offset
                if topk_local >= 0 and topk_local < seqlen_k:
                    self._copy_row(
                        mK,
                        sK_slice,
                        row,
                        idx_in_group,
                        copy_atom,
                        thr_copy,
                        topk_local,
                        batch_idx,
                        k_offset=k_offset,
                    )

    @cute.jit
    def _copy_row(
        self,
        mK: cute.Tensor,
        sK_slice: cute.Tensor,
        row: Int32,
        idx_in_group: Int32,
        copy_atom: cute.CopyAtom,
        thr_copy: cute.TiledCopy,
        topk_idx: Int32,
        batch_idx: Int32,
        k_offset: int = 0,
    ):
        """Copy one (partial) K row from gmem to smem via cp.async (8 threads per row).

        When k_block_size < head_dim_padded, k_offset selects which k_block_size
        slice of the row to copy. sK_slice has shape (n_block_size, k_block_size).
        """
        gK_row_raw = mK[topk_idx, None, batch_idx]
        gK_row_offset = gK_row_raw.iterator + k_offset
        gK_row = cute.make_tensor(
            cute.make_ptr(
                self.k_dtype,
                gK_row_offset.llvm_ptr,
                cute.AddressSpace.gmem,
                assumed_align=16,
            ),
            cute.make_layout((self.k_block_size,)),
        )
        gK_chunks = cute.flat_divide(gK_row, (8,))
        sK_row = sK_slice[row, None]
        sK_chunks = cute.flat_divide(sK_row, (8,))
        for tile in cutlass.range_constexpr(self.k_block_size // 64):
            chunk_idx = tile * 8 + idx_in_group
            g_chunk = gK_chunks[None, chunk_idx]
            s_chunk = sK_chunks[None, chunk_idx]
            tSg = thr_copy.partition_S(g_chunk)
            tSs = thr_copy.partition_D(s_chunk)
            cute.copy(copy_atom, tSg, tSs)

    @cute.jit
    def _zero_row(self, sK_slice: cute.Tensor, row: Int32, idx_in_group: Int32):
        """Zero-fill one K row in smem, cooperative across 8 threads in a group."""
        sK_row = sK_slice[row, None]
        sK_chunks = cute.flat_divide(sK_row, (8,))
        for tile in cutlass.range_constexpr(self.k_block_size // 64):
            chunk_idx = tile * 8 + idx_in_group
            sK_chunks[None, chunk_idx].fill(0)

    # =========================================================================
    # Cross-warp reduce helpers (epilogue warpgroup)
    # =========================================================================
    @cute.jit
    def _intra_inter_warp_reduce_max(
        self,
        sScoreAll,
        reduce_sync_mbar_ptr,
        reduce_sync_phase,
        warp_id_in_wg,
        local_value,
    ):
        """Reduce local max across 4 warps via smem scratch + mbarrier sync."""
        warp_max = cute.arch.warp_redux_sync(local_value, "fmax")
        with cute.arch.elect_one():
            sScoreAll[warp_id_in_wg] = warp_max
        cute.arch.fence_view_async_shared()
        cute.arch.mbarrier_arrive(reduce_sync_mbar_ptr)
        cute.arch.mbarrier_wait(reduce_sync_mbar_ptr, reduce_sync_phase)
        reduce_sync_phase = reduce_sync_phase ^ 1

        global_max = sScoreAll[0]
        for wi in cutlass.range_constexpr(self.num_warps_in_epi_wg - 1):
            v = sScoreAll[wi + 1]
            global_max = v if v > global_max else global_max

        return global_max, reduce_sync_phase

    @cute.jit
    def _intra_inter_warp_reduce_sum(
        self,
        sScoreAll,
        reduce_sync_mbar_ptr,
        reduce_sync_phase,
        warp_id_in_wg,
        local_value,
    ):
        """Reduce local sum across 4 warps via smem scratch + mbarrier sync."""
        warp_sum = cute.arch.warp_reduction_sum(local_value)
        with cute.arch.elect_one():
            sScoreAll[warp_id_in_wg] = warp_sum
        cute.arch.fence_view_async_shared()
        cute.arch.mbarrier_arrive(reduce_sync_mbar_ptr)
        cute.arch.mbarrier_wait(reduce_sync_mbar_ptr, reduce_sync_phase)
        reduce_sync_phase = reduce_sync_phase ^ 1

        global_sum = sScoreAll[0]
        for wi in cutlass.range_constexpr(self.num_warps_in_epi_wg - 1):
            global_sum = global_sum + sScoreAll[wi + 1]

        return global_sum, reduce_sync_phase

    @cute.jit
    def _inter_warp_sync_sum(
        self,
        sScoreAll,
        reduce_sync_mbar_ptr,
        reduce_sync_phase,
        warp_id_in_wg,
        warp_sum,
    ):
        """Cross-warp sum sync with pre-computed warp-level sum (skip warp_reduction_sum)."""
        with cute.arch.elect_one():
            sScoreAll[warp_id_in_wg] = warp_sum
        cute.arch.fence_view_async_shared()
        cute.arch.mbarrier_arrive(reduce_sync_mbar_ptr)
        cute.arch.mbarrier_wait(reduce_sync_mbar_ptr, reduce_sync_phase)
        reduce_sync_phase = reduce_sync_phase ^ 1

        global_sum = sScoreAll[0]
        for wi in cutlass.range_constexpr(self.num_warps_in_epi_wg - 1):
            global_sum = global_sum + sScoreAll[wi + 1]

        return global_sum, reduce_sync_phase

    # =========================================================================
    # Epilogue: indexer mode — ReLU(QK) * W, head reduce, softmax
    # =========================================================================
    @cute.jit
    def _epilogue_indexer(
        self,
        tiled_mma_qk,
        tStS_ref,
        mPerHead,
        sW,
        sScoreAll,
        S_mbar_ptr,
        reduce_sync_mbar_ptr,
        sTopkIdx,
        mTopkIdx,
        mOut,
        mTopkLength,
        m_block,
        batch_idx,
        tidx,
        s_full_phase,
        reduce_phase,
        softmax_scale,
        per_head_offset=None,
        topk_idx_offset=None,
    ):
        """Epilogue: TMEM->RF, ReLU, *W, head reduce -> scores -> softmax -> predict.

        1. n_block loop: ReLU(QK)*W head reduce, accumulate scores in rmem
        2. Post-loop: mask invalid positions, compute softmax, write predict to gmem

        Non-compact: W loaded directly from GMEM into registers (bypass SMEM
        to avoid LDS bandwidth contention with K scatter-gather loading).
        Compact: W loaded from SMEM (original path, preserves compact perf).
        """
        tidx_wg = tidx % self.WARPGROUP_SIZE

        tmem_load_atom = cute.make_copy_atom(
            tcgen05.copy.Ld32x32bOp(tcgen05.copy.Repetition(8)),
            Float32,
        )
        thr_tmem_load = tcgen05.make_tmem_copy(tmem_load_atom, tStS_ref).get_slice(tidx_wg)

        # Coordinate tensor: MMA partition -> remap to TMEM copy destination layout
        thr_mma = tiled_mma_qk.get_slice(tidx_wg)
        cS = cute.make_identity_tensor(self.mma_tiler_qk[:2])
        tScS = thr_tmem_load.partition_D(thr_mma.partition_C(cS))

        tSrS_shape = thr_tmem_load.partition_D(cute.make_identity_tensor(tStS_ref.shape)).shape
        tSrS = cute.make_rmem_tensor(tSrS_shape, Float32)

        sW_off = Int32(0) if per_head_offset is None else per_head_offset
        sTopkIdx_off = Int32(0) if topk_idx_offset is None else topk_idx_offset
        qhpkv = self.qhead_per_kvhead

        W_ILP = 4
        rW_all = cute.make_rmem_tensor((self.m_block_size,), self.per_head_dtype)
        rW_all_f32 = cute.recast_tensor(rW_all, Float32)

        # Load W from GMEM directly (bypass SMEM, reduce LDS contention
        # with K scatter-gather cp.async writes to SMEM).
        q_token_idx_w = m_block
        gW_raw = mPerHead[q_token_idx_w, None, batch_idx]
        gW_f32_ptr = cute.make_ptr(
            Float32,
            gW_raw.iterator.llvm_ptr,
            cute.AddressSpace.gmem,
            assumed_align=16,
        )
        W_src_f32 = cute.make_tensor(
            gW_f32_ptr,
            cute.make_layout((self.m_block_size // 2,)),
        )

        kv_offset = tScS[0][0]
        warp_id_in_wg = tidx_wg // self.WARP_SIZE

        # ---- Phase 1: n_block loop — head reduce, hold scores in rmem ----
        # Process all n_blocks unconditionally (no block-level skip).
        # MMA produces all blocks with static bounds; invalid scores masked in Phase 2.
        rScores = [Float32(0.0) for _ in range(self.num_n_blocks)]

        q_token_idx = m_block

        if const_expr(self.have_topk_length):
            topK_epi = mTopkLength[q_token_idx, batch_idx]

        for _ri in cutlass.range_constexpr(self.num_n_blocks):
            n_blk = self.num_n_blocks - 1 - _ri
            _slot = const_expr(n_blk % self.num_tmem_slots)
            cute.arch.mbarrier_wait(S_mbar_ptr + 2 * _slot, s_full_phase)
            if const_expr(_ri == 0):
                cute.autovec_copy(W_src_f32, rW_all_f32)
            tmem_ptr_cur = cute.make_ptr(
                Float32,
                _slot * self.tmem_s_stride,
                mem_space=cute.AddressSpace.tmem,
                assumed_align=16,
            )
            tStS_cur = cute.make_tensor(tmem_ptr_cur, tStS_ref.layout)
            tStS_t2r_cur = thr_tmem_load.partition_S(tStS_cur)

            cute.copy(thr_tmem_load, tStS_t2r_cur, tSrS)
            cute.arch.fence_view_async_tmem_load()
            cute.arch.mbarrier_arrive(S_mbar_ptr + 2 * _slot + 1)
            if const_expr(_slot == 0):
                s_full_phase ^= 1
            local_sum = (Float32(0.0), Float32(0.0))
            for ho in cutlass.range_constexpr(qhpkv // 2 // W_ILP):
                for ci in cutlass.range_constexpr(W_ILP):
                    idx0 = (ho * W_ILP + ci) * 2
                    idx1 = idx0 + 1
                    w_pair = (Float32(rW_all[idx0]), Float32(rW_all[idx1]))

                    val0 = tSrS[idx0]
                    val0 = val0 if val0 > Float32(0.0) else Float32(0.0)
                    val1 = tSrS[idx1]
                    val1 = val1 if val1 > Float32(0.0) else Float32(0.0)

                    prod = mul_packed_f32x2((val0, val1), w_pair)
                    local_sum = add_packed_f32x2(local_sum, prod)

            rScores[n_blk] = (local_sum[0] + local_sum[1]) * Float32(softmax_scale)

        # ---- Phase 2: softmax (scores in rmem, only 4 FP32 smem for cross-warp) ----
        # Mask invalid positions + local max
        local_max = -Float32.inf
        for ei in cutlass.range_constexpr(self.num_n_blocks):
            pos = kv_offset + ei * self.n_block_size
            if const_expr(self.have_topk_length):
                if pos >= topK_epi:
                    rScores[ei] = -Float32.inf
            else:
                if const_expr(self.topk_in_smem):
                    topk_idx = Int32(sTopkIdx[sTopkIdx_off + pos])
                else:
                    topk_idx = Int32(mTopkIdx[q_token_idx, pos, batch_idx])
                if topk_idx < 0:
                    rScores[ei] = -Float32.inf
            if rScores[ei] > local_max:
                local_max = rScores[ei]

        global_max, reduce_phase = self._intra_inter_warp_reduce_max(
            sScoreAll,
            reduce_sync_mbar_ptr,
            reduce_phase,
            warp_id_in_wg,
            local_max,
        )

        if global_max > Float32(-1e30):
            # Compute exp(score - max) and local sum
            log2_e = Float32(math.log2(math.e))
            local_exp_sum = Float32(0.0)
            for ei in cutlass.range_constexpr(self.num_n_blocks):
                exp_val = cute.math.exp2((rScores[ei] - global_max) * log2_e)
                rScores[ei] = exp_val
                local_exp_sum = local_exp_sum + exp_val

            sScoreAll_sum = cute.make_tensor(
                sScoreAll.iterator + self.num_warps_in_epi_wg,
                cute.make_layout((self.num_warps_in_epi_wg,), stride=(1,)),
            )
            global_sum, reduce_phase = self._intra_inter_warp_reduce_sum(
                sScoreAll_sum,
                reduce_sync_mbar_ptr,
                reduce_phase,
                warp_id_in_wg,
                local_exp_sum,
            )

            inv_sum = Float32(1.0) / global_sum
            for ei in cutlass.range_constexpr(self.num_n_blocks):
                pos = kv_offset + ei * self.n_block_size
                mOut[q_token_idx, pos, batch_idx] = rScores[ei] * inv_sum
        else:
            for ei in cutlass.range_constexpr(self.num_n_blocks):
                pos = kv_offset + ei * self.n_block_size
                mOut[q_token_idx, pos, batch_idx] = Float32(0.0)
            # 1 dummy sync to match the if-branch's sum reduce sync
            cute.arch.mbarrier_arrive(reduce_sync_mbar_ptr)
            cute.arch.mbarrier_wait(reduce_sync_mbar_ptr, reduce_phase)
            reduce_phase = reduce_phase ^ 1

        return s_full_phase, reduce_phase

    # =========================================================================
    # Epilogue: attention mode, n_block_size>=128 — Ld32x32bOp path
    # =========================================================================
    @cute.jit
    def _epilogue_attention_n128(
        self,
        tiled_mma_qk,
        tStS_ref,
        sLSE,
        sScoreAll,
        S_mbar_ptr,
        reduce_sync_mbar_ptr,
        sTopkIdx,
        mTopkIdx,
        mOut,
        mTopkLength,
        m_block,
        batch_idx,
        tidx,
        s_full_phase,
        reduce_phase,
        softmax_scale,
        per_head_offset=None,
        topk_idx_offset=None,
    ):
        """Epilogue (attention, n_block_size>=128): TMEM->RF, exp2(S*scale - LSE), head reduce -> L1-norm.

        With n_block_size>=128 (MMA M>=128), TMEM output occupies contiguous
        lanes, so we use Ld32x32bOp. Each thread owns all m_block_size heads;
        no shfl_xor or is_active gating needed.
        """
        tidx_wg = tidx % self.WARPGROUP_SIZE

        tmem_load_atom = cute.make_copy_atom(
            tcgen05.copy.Ld32x32bOp(tcgen05.copy.Repetition(8)),
            Float32,
        )
        thr_tmem_load = tcgen05.make_tmem_copy(tmem_load_atom, tStS_ref).get_slice(tidx_wg)

        thr_mma = tiled_mma_qk.get_slice(tidx_wg)
        cS = cute.make_identity_tensor(self.mma_tiler_qk[:2])
        tScS = thr_tmem_load.partition_D(thr_mma.partition_C(cS))

        tSrS_shape = thr_tmem_load.partition_D(cute.make_identity_tensor(tStS_ref.shape)).shape
        tSrS = cute.make_rmem_tensor(tSrS_shape, Float32)

        sLSE_off = Int32(0) if per_head_offset is None else per_head_offset
        sTopkIdx_off = Int32(0) if topk_idx_offset is None else topk_idx_offset
        qhpkv = self.qhead_per_kvhead

        LSE_ILP = 4
        sLSE_f32_ptr = cute.make_ptr(
            Float32,
            (sLSE.iterator + sLSE_off).llvm_ptr,
            cute.AddressSpace.smem,
            assumed_align=16,
        )
        sLSE_1d = cute.make_tensor(
            sLSE_f32_ptr,
            cute.make_layout((self.m_block_size,)),
        )
        rLSE_all = cute.make_rmem_tensor((self.m_block_size,), Float32)

        kv_offset = tScS[0][0]
        warp_id_in_wg = tidx_wg // self.WARP_SIZE

        log2_e = Float32(math.log2(math.e))
        scale_log2_e = Float32(softmax_scale) * log2_e

        q_token_idx = m_block

        # ---- Phase 1: n_block loop — softmax recovery + head reduce ----
        rScores = [Float32(0.0) for _ in range(self.num_n_blocks)]
        warp_sum_acc = Float32(0.0)

        if const_expr(self.have_topk_length):
            topK_epi = mTopkLength[q_token_idx, batch_idx]
            n_block_max_epi = (topK_epi + self.n_block_size - 1) // self.n_block_size

        for _ri in cutlass.range_constexpr(self.num_n_blocks):
            n_blk = self.num_n_blocks - 1 - _ri
            _slot = const_expr(n_blk % self.num_tmem_slots)
            cute.arch.mbarrier_wait(S_mbar_ptr + 2 * _slot, s_full_phase)
            if const_expr(_ri == 0):
                cute.autovec_copy(sLSE_1d, rLSE_all)
            if const_expr(not self.have_topk_length) or n_blk < n_block_max_epi:
                tmem_ptr_cur = cute.make_ptr(
                    Float32,
                    _slot * self.tmem_s_stride,
                    mem_space=cute.AddressSpace.tmem,
                    assumed_align=16,
                )
                tStS_cur = cute.make_tensor(tmem_ptr_cur, tStS_ref.layout)
                tStS_t2r_cur = thr_tmem_load.partition_S(tStS_cur)

                cute.copy(thr_tmem_load, tStS_t2r_cur, tSrS)
                cute.arch.fence_view_async_tmem_load()
            cute.arch.mbarrier_arrive(S_mbar_ptr + 2 * _slot + 1)
            if const_expr(_slot == 0):
                s_full_phase ^= 1
            if const_expr(not self.have_topk_length) or n_blk < n_block_max_epi:
                local_sum = (Float32(0.0), Float32(0.0))
                for ho in cutlass.range_constexpr(qhpkv // 2 // LSE_ILP):
                    for ci in cutlass.range_constexpr(LSE_ILP):
                        idx0 = (ho * LSE_ILP + ci) * 2
                        idx1 = idx0 + 1
                        lse_pair = (rLSE_all[idx0], rLSE_all[idx1])

                        val0 = tSrS[idx0]
                        val1 = tSrS[idx1]

                        val0, val1 = fma_packed_f32x2(
                            (val0, val1),
                            (scale_log2_e, scale_log2_e),
                            lse_pair,
                        )
                        val0 = cute.math.exp2(val0)
                        val1 = cute.math.exp2(val1)

                        local_sum = add_packed_f32x2(local_sum, (val0, val1))

                rScores[n_blk] = local_sum[0] + local_sum[1]

                pos = kv_offset + n_blk * self.n_block_size
                if const_expr(self.have_topk_length):
                    if pos >= topK_epi:
                        rScores[n_blk] = Float32(0.0)
                else:
                    if const_expr(self.topk_in_smem):
                        topk_idx = Int32(sTopkIdx[sTopkIdx_off + pos])
                    else:
                        topk_idx = Int32(mTopkIdx[q_token_idx, pos, batch_idx])
                    if topk_idx < 0:
                        rScores[n_blk] = Float32(0.0)
                warp_sum_acc = warp_sum_acc + cute.arch.warp_reduction_sum(rScores[n_blk])

        # ---- Phase 2: L1-norm (scores already non-negative from exp) ----
        global_sum, reduce_phase = self._inter_warp_sync_sum(
            sScoreAll,
            reduce_sync_mbar_ptr,
            reduce_phase,
            warp_id_in_wg,
            warp_sum_acc,
        )

        if global_sum > Float32(1e-10):
            inv_sum = Float32(1.0) / global_sum
            for ei in cutlass.range_constexpr(self.num_n_blocks):
                pos = kv_offset + ei * self.n_block_size
                mOut[q_token_idx, pos, batch_idx] = rScores[ei] * inv_sum
        else:
            for ei in cutlass.range_constexpr(self.num_n_blocks):
                pos = kv_offset + ei * self.n_block_size
                mOut[q_token_idx, pos, batch_idx] = Float32(0.0)

        return s_full_phase, reduce_phase

    # =========================================================================
    # Epilogue: attention mode, n_block_size<128 — Ld16x64bOp path
    # =========================================================================
    @cute.jit
    def _epilogue_attention(
        self,
        tiled_mma_qk,
        tStS_ref,
        sLSE,
        sScoreAll,
        S_mbar_ptr,
        reduce_sync_mbar_ptr,
        sTopkIdx,
        mTopkIdx,
        mOut,
        mTopkLength,
        m_block,
        batch_idx,
        tidx,
        s_full_phase,
        reduce_phase,
        softmax_scale,
        per_head_offset=None,
        topk_idx_offset=None,
    ):
        """Epilogue (attention mode): TMEM->RF, exp2(S*scale - LSE), head reduce -> L1-norm.

        With n_block_size=64 (MMA M=64), TMEM output occupies non-contiguous
        lanes (0-15, 32-47, 64-79, 96-111), so we use Ld16x64bOp instead of
        Ld32x32bOp. Per PTX Figure 184, each lane is shared by Thread i and
        Thread i^2 (e.g. T0+T2 → lane 0, T1+T3 → lane 8). Each thread holds
        m_block_size/2 heads; a single shfl_xor(2) combines the partial sums.
        """
        tidx_wg = tidx % self.WARPGROUP_SIZE

        tmem_load_atom = cute.make_copy_atom(
            tcgen05.copy.Ld16x64bOp(tcgen05.copy.Repetition(self.m_block_size // 2)),
            Float32,
        )
        thr_tmem_load = tcgen05.make_tmem_copy(tmem_load_atom, tStS_ref).get_slice(tidx_wg)

        thr_mma = tiled_mma_qk.get_slice(tidx_wg)
        cS = cute.make_identity_tensor(self.mma_tiler_qk[:2])
        tScS = thr_tmem_load.partition_D(thr_mma.partition_C(cS))

        tSrS_shape = thr_tmem_load.partition_D(cute.make_identity_tensor(tStS_ref.shape)).shape
        tSrS = cute.make_rmem_tensor(tSrS_shape, Float32)

        sLSE_off = Int32(0) if per_head_offset is None else per_head_offset
        sTopkIdx_off = Int32(0) if topk_idx_offset is None else topk_idx_offset

        # Number of heads each thread processes (half of m_block_size)
        num_heads_per_thr = self.m_block_size // 2
        LSE_ILP = 4

        sLSE_f32_ptr = cute.make_ptr(
            Float32,
            (sLSE.iterator + sLSE_off).llvm_ptr,
            cute.AddressSpace.smem,
            assumed_align=16,
        )

        # Broadcast SMEM LSE view: shape (n_block_size, m_block_size) with
        # stride=(0,1) so every M-row sees the same per-head LSE values.
        # Partitioning with the same MMA+TMEM_copy chain as tSrS ensures
        # tSrLSE[i] pairs with tSrS[i] regardless of fragment layout.
        sLSE_2d = cute.make_tensor(
            sLSE_f32_ptr,
            cute.make_layout((self.n_block_size, self.m_block_size), stride=(0, 1)),
        )
        tSsLSE = thr_tmem_load.partition_D(thr_mma.partition_C(sLSE_2d))
        tSrLSE = cute.make_rmem_tensor(tSsLSE.shape, Float32)

        kv_offset = tScS[0][0]
        warp_id_in_wg = tidx_wg // self.WARP_SIZE
        # Thread i and Thread i^2 share a lane; only one should contribute
        # to warp-level sums / write output.  (tidx_wg % 4) < 2 selects
        # exactly one thread per lane (T0,T1,T4,T5,...  not T2,T3,T6,T7,...).
        is_active = (tidx_wg % 4) < 2

        log2_e = Float32(math.log2(math.e))
        scale_log2_e = Float32(softmax_scale) * log2_e

        q_token_idx = m_block

        # ---- Phase 1: n_block loop — softmax recovery + head reduce, hold scores in rmem ----
        rScores = [Float32(0.0) for _ in range(self.num_n_blocks)]
        warp_sum_acc = Float32(0.0)

        if const_expr(self.have_topk_length):
            topK_epi = mTopkLength[q_token_idx, batch_idx]
            n_block_max_epi = (topK_epi + self.n_block_size - 1) // self.n_block_size

        for _ri in cutlass.range_constexpr(self.num_n_blocks):
            n_blk = self.num_n_blocks - 1 - _ri
            _slot = const_expr(n_blk % self.num_tmem_slots)
            cute.arch.mbarrier_wait(S_mbar_ptr + 2 * _slot, s_full_phase)
            if const_expr(_ri == 0):
                cute.autovec_copy(tSsLSE, tSrLSE)
            if const_expr(not self.have_topk_length) or n_blk < n_block_max_epi:
                tmem_ptr_cur = cute.make_ptr(
                    Float32,
                    _slot * self.tmem_s_stride,
                    mem_space=cute.AddressSpace.tmem,
                    assumed_align=16,
                )
                tStS_cur = cute.make_tensor(tmem_ptr_cur, tStS_ref.layout)
                tStS_t2r_cur = thr_tmem_load.partition_S(tStS_cur)

                cute.copy(thr_tmem_load, tStS_t2r_cur, tSrS)
                cute.arch.fence_view_async_tmem_load()
            cute.arch.mbarrier_arrive(S_mbar_ptr + 2 * _slot + 1)
            if const_expr(_slot == 0):
                s_full_phase ^= 1
            if const_expr(not self.have_topk_length) or n_blk < n_block_max_epi:
                local_sum = (Float32(0.0), Float32(0.0))
                for ho in cutlass.range_constexpr(num_heads_per_thr // 2 // LSE_ILP):
                    for ci in cutlass.range_constexpr(LSE_ILP):
                        idx0 = (ho * LSE_ILP + ci) * 2
                        idx1 = idx0 + 1
                        lse_pair = (tSrLSE[idx0], tSrLSE[idx1])

                        val0 = tSrS[idx0]
                        val1 = tSrS[idx1]

                        val0, val1 = fma_packed_f32x2(
                            (val0, val1),
                            (scale_log2_e, scale_log2_e),
                            lse_pair,
                        )
                        val0 = cute.math.exp2(val0)
                        val1 = cute.math.exp2(val1)

                        local_sum = add_packed_f32x2(local_sum, (val0, val1))

                partial_score = local_sum[0] + local_sum[1]

                # Mask invalid KV positions at partial level
                pos = kv_offset + n_blk * self.n_block_size
                if const_expr(self.have_topk_length):
                    if pos >= topK_epi:
                        partial_score = Float32(0.0)
                else:
                    if const_expr(self.topk_in_smem):
                        topk_idx = Int32(sTopkIdx[sTopkIdx_off + pos])
                    else:
                        topk_idx = Int32(mTopkIdx[q_token_idx, pos, batch_idx])
                    if topk_idx < 0:
                        partial_score = Float32(0.0)

                warp_sum_acc = warp_sum_acc + cute.arch.warp_reduction_sum(partial_score)

                partner_partial = cute.arch.shuffle_sync_bfly(
                    partial_score,
                    offset=2,
                    mask=-1,
                    mask_and_clamp=31,
                )
                rScores[n_blk] = partial_score + partner_partial

        # ---- Phase 2: L1-norm (scores already non-negative from exp) ----
        global_sum, reduce_phase = self._inter_warp_sync_sum(
            sScoreAll,
            reduce_sync_mbar_ptr,
            reduce_phase,
            warp_id_in_wg,
            warp_sum_acc,
        )

        if is_active:
            if global_sum > Float32(1e-10):
                inv_sum = Float32(1.0) / global_sum
                for ei in cutlass.range_constexpr(self.num_n_blocks):
                    pos = kv_offset + ei * self.n_block_size
                    mOut[q_token_idx, pos, batch_idx] = rScores[ei] * inv_sum
            else:
                for ei in cutlass.range_constexpr(self.num_n_blocks):
                    pos = kv_offset + ei * self.n_block_size
                    mOut[q_token_idx, pos, batch_idx] = Float32(0.0)

        return s_full_phase, reduce_phase
